from dataclasses import dataclass
from re import compile
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Set,
)

from packaging.version import Version

from galaxy.tool_util.parser.interface import (
    TestCollectionDef,
    ToolSource,
    ToolSourceTest,
    ToolSourceTestInput,
    ToolSourceTestInputs,
    xml_data_input_to_json,
    XmlTestCollectionDefDict,
)
from galaxy.tool_util.parser.util import multiple_select_value_split
from galaxy.tool_util_models.parameters import (
    BooleanParameterModel,
    ConditionalParameterModel,
    ConditionalWhen,
    DataCollectionParameterModel,
    DataColumnParameterModel,
    DataParameterModel,
    FloatParameterModel,
    GenomeBuildParameterModel,
    GroupTagParameterModel,
    IntegerParameterModel,
    RepeatParameterModel,
    SectionParameterModel,
    ToolParameterT,
)
from galaxy.util import asbool
from .factory import input_models_for_tool_source
from .state import TestCaseToolState
from .visitor import (
    flat_state_path,
    repeat_inputs_to_array,
    validate_explicit_conditional_test_value,
)

INTEGER_STR_PATTERN = compile(r"^(\d+)$")
INTEGERS_STR_PATTERN = compile(r"^(\d+)(\s*,\s*(\d+))*$")
COLUMN_NAME_STR_PATTERN = compile(r"^c(\d+): .*$")
# In an effort to squeeze all the ambiguity out of test cases - at some point Marius and John
# agree tools should be using value_json for typed inputs to parameters but John has come around on
# this now that we're validating the parameters as a whole on load. The models are ensuring only
# unambigious test cases are being loaded.
WARN_ON_UNTYPED_XML_STRINGS = False


@dataclass
class TestCaseStateAndWarnings:
    tool_state: TestCaseToolState
    warnings: List[str]
    unhandled_inputs: List[str]


@dataclass
class TestCaseStateValidationResult:
    tool_state: TestCaseToolState
    warnings: List[str]
    validation_error: Optional[Exception]
    tool_parameter_bundle: List[ToolParameterT]
    profile: str

    def to_dict(self):
        tool_state_json = self.tool_state.input_state
        warnings = self.warnings
        validation_error = str(self.validation_error) if self.validation_error else None
        return {
            "tool_state": tool_state_json,
            "warnings": warnings,
            "validation_error": validation_error,
            "validated_with_profile": self.profile,
        }


def legacy_from_string(parameter: ToolParameterT, value: Optional[Any], warnings: List[str], profile: str) -> Any:
    """Convert string values in XML test cases into typed variants.

    This should only be used when parsing XML test cases into a TestCaseToolState object.
    We have to maintain backward compatibility on these for older Galaxy tool profile versions.
    """
    result_value: Any = value
    if isinstance(value, str):
        if isinstance(parameter, (IntegerParameterModel,)):
            if WARN_ON_UNTYPED_XML_STRINGS:
                warnings.append(
                    f"Implicitly converted {parameter.name} to an integer from a string value, please use 'value_json' to define this test input parameter value instead."
                )
            result_value = int(value)
        elif isinstance(parameter, (FloatParameterModel,)):
            if WARN_ON_UNTYPED_XML_STRINGS:
                warnings.append(
                    f"Implicitly converted {parameter.name} to a floating point number from a string value, please use 'value_json' to define this test input parameter value instead."
                )
            result_value = float(value)
        elif isinstance(parameter, (BooleanParameterModel,)):
            if WARN_ON_UNTYPED_XML_STRINGS:
                warnings.append(
                    f"Implicitly converted {parameter.name} to a boolean from a string value, please use 'value_json' to define this test input parameter value instead."
                )
            try:
                result_value = asbool(value)
            except ValueError:
                warnings.append(
                    "Likely using deprected truevalue/falsevalue in tool parameter - switch to 'true' or 'false'"
                )
        elif isinstance(parameter, (GroupTagParameterModel,)):
            if parameter.multiple:
                result_value = multiple_select_value_split(value)
        elif isinstance(parameter, (GenomeBuildParameterModel,)):
            if parameter.multiple:
                result_value = multiple_select_value_split(value)
        elif isinstance(parameter, (DataColumnParameterModel,)):
            if parameter.multiple:
                integers_match = INTEGER_STR_PATTERN.match(value)
                if integers_match:
                    if WARN_ON_UNTYPED_XML_STRINGS:
                        warnings.append(
                            f"Implicitly converted {parameter.name} to a column index integer from a string value, please use 'value_json' to define this test input parameter value instead."
                        )
                    result_value = [int(v.strip()) for v in value.split(",")]
            else:
                integer_match = INTEGER_STR_PATTERN.match(value)
                if integer_match:
                    if WARN_ON_UNTYPED_XML_STRINGS:
                        warnings.append(
                            f"Implicitly converted {parameter.name} to a column index integer from a string value, please use 'value_json' to define this test input parameter value instead."
                        )
                    result_value = int(value)
                elif Version(profile) < Version("24.2"):
                    # allow this for older tools but new tools will just require the integer index
                    warnings.append(
                        f"Using column names as test case values is deprecated, please adjust {parameter.name} to just use an integer column index."
                    )
                    column_name_value_match = COLUMN_NAME_STR_PATTERN.match(value)
                    if column_name_value_match:
                        column_part = column_name_value_match.group(1)
                        result_value = int(column_part)
    return result_value


def test_case_state(
    test_dict: ToolSourceTest,
    tool_parameter_bundle: List[ToolParameterT],
    profile: str,
    validate: bool = True,
    name: Optional[str] = None,
) -> TestCaseStateAndWarnings:
    warnings: List[str] = []
    inputs: ToolSourceTestInputs = test_dict["inputs"]
    unhandled_inputs = []
    state: Dict[str, Any] = {}

    handled_inputs = _merge_level_into_state(tool_parameter_bundle, inputs, state, profile, warnings, None)

    for test_input in inputs:
        input_name = test_input["name"]
        if input_name not in handled_inputs:
            unhandled_inputs.append(input_name)

    tool_state = TestCaseToolState(state)
    if validate:
        tool_state.validate(tool_parameter_bundle, name=name)
        for input_name in unhandled_inputs:
            raise Exception(f"Invalid parameter name found {input_name}")
    return TestCaseStateAndWarnings(tool_state, warnings, unhandled_inputs)


def test_case_validation(
    test_dict: ToolSourceTest, tool_parameter_bundle: List[ToolParameterT], profile: str, name: Optional[str] = None
) -> TestCaseStateValidationResult:
    test_case_state_and_warnings = test_case_state(test_dict, tool_parameter_bundle, profile, validate=False)
    exception: Optional[Exception] = None
    try:
        test_case_state_and_warnings.tool_state.validate(tool_parameter_bundle, name=name)
        for input_name in test_case_state_and_warnings.unhandled_inputs:
            raise Exception(f"Invalid parameter name found {input_name}")
    except Exception as e:
        exception = e
    return TestCaseStateValidationResult(
        test_case_state_and_warnings.tool_state,
        test_case_state_and_warnings.warnings,
        exception,
        tool_parameter_bundle,
        profile,
    )


def _merge_level_into_state(
    tool_inputs: List[ToolParameterT],
    inputs: ToolSourceTestInputs,
    state_at_level: dict,
    profile: str,
    warnings: List[str],
    prefix: Optional[str],
) -> Set[str]:
    handled_inputs: Set[str] = set()
    for tool_input in tool_inputs:
        handled_inputs.update(_merge_into_state(tool_input, inputs, state_at_level, profile, warnings, prefix))

    return handled_inputs


def _inputs_as_dict(inputs: ToolSourceTestInputs) -> Dict[str, ToolSourceTestInput]:
    as_dict: Dict[str, ToolSourceTestInput] = {}
    for input in inputs:
        as_dict[input["name"]] = input

    return as_dict


def _merge_into_state(
    tool_input: ToolParameterT,
    inputs: ToolSourceTestInputs,
    state_at_level: dict,
    profile: str,
    warnings: List[str],
    prefix: Optional[str],
) -> Set[str]:
    handled_inputs = set()

    input_name = tool_input.name
    state_path = flat_state_path(input_name, prefix)
    handled_inputs.add(state_path)

    if isinstance(tool_input, (ConditionalParameterModel,)):
        conditional_state = state_at_level.get(input_name, {})
        if input_name not in state_at_level:
            state_at_level[input_name] = conditional_state

        when: ConditionalWhen = _select_which_when(tool_input, conditional_state, inputs, state_path)
        test_parameter = tool_input.test_parameter
        handled_inputs.update(
            _merge_into_state(test_parameter, inputs, conditional_state, profile, warnings, state_path)
        )
        handled_inputs.update(
            _merge_level_into_state(when.parameters, inputs, conditional_state, profile, warnings, state_path)
        )
    elif isinstance(tool_input, (RepeatParameterModel,)):
        repeat_state_array = state_at_level.get(input_name, [])
        if input_name not in state_at_level:
            state_at_level[input_name] = repeat_state_array

        repeat_instance_inputs = repeat_inputs_to_array(state_path, _inputs_as_dict(inputs))
        if tool_input.min is not None:
            while len(repeat_instance_inputs) < tool_input.min:
                repeat_instance_inputs.append({})
        for i, _ in enumerate(repeat_instance_inputs):
            while len(repeat_state_array) <= i:
                repeat_state_array.append({})

            repeat_instance_prefix = f"{state_path}_{i}"
            handled_inputs.update(
                _merge_level_into_state(
                    tool_input.parameters, inputs, repeat_state_array[i], profile, warnings, repeat_instance_prefix
                )
            )
    elif isinstance(tool_input, (SectionParameterModel,)):
        section_state = state_at_level.get(input_name, {})
        if input_name not in state_at_level:
            state_at_level[input_name] = section_state

        handled_inputs.update(
            _merge_level_into_state(tool_input.parameters, inputs, section_state, profile, warnings, state_path)
        )
    else:
        test_input = _input_for(state_path, inputs)
        if test_input is not None:
            input_value: Any
            if isinstance(tool_input, (DataCollectionParameterModel,)):
                input_value = TestCollectionDef.from_dict(
                    cast(XmlTestCollectionDefDict, test_input.get("attributes", {}).get("collection"))
                ).test_format_to_dict()
            elif isinstance(tool_input, (DataParameterModel,)):
                if tool_input.multiple:
                    value = test_input["value"]
                    input_value_list = []
                    if value:
                        test_input_values = cast(str, value).split(",")
                        for test_input_value in test_input_values:
                            instance_test_input = test_input.copy()
                            instance_test_input["value"] = test_input_value
                            input_value_json = xml_data_input_to_json(instance_test_input)
                            input_value_list.append(input_value_json)
                    input_value = input_value_list
                else:
                    input_value = xml_data_input_to_json(test_input)
            else:
                input_value = test_input["value"]
                input_value = legacy_from_string(tool_input, input_value, warnings, profile)

            state_at_level[input_name] = input_value

    return handled_inputs


def _select_which_when(
    conditional: ConditionalParameterModel, state: dict, inputs: ToolSourceTestInputs, prefix: str
) -> ConditionalWhen:
    test_parameter = conditional.test_parameter
    is_boolean = test_parameter.parameter_type == "gx_boolean"
    test_parameter_name = test_parameter.name
    test_parameter_flat_path = flat_state_path(test_parameter_name, prefix)

    test_input = _input_for(test_parameter_flat_path, inputs)
    explicit_test_value = test_input["value"] if test_input else None
    if is_boolean and isinstance(explicit_test_value, str):
        explicit_test_value = asbool(explicit_test_value)
    test_value = validate_explicit_conditional_test_value(test_parameter_name, explicit_test_value)
    for when in conditional.whens:
        if test_value is None and when.is_default_when:
            return when
        elif test_value == when.discriminator:
            return when
    else:
        raise Exception(f"Invalid conditional test value ({explicit_test_value}) for parameter ({test_parameter_name})")


def _input_for(flat_state_path: str, inputs: ToolSourceTestInputs) -> Optional[ToolSourceTestInput]:
    for input in inputs:
        if input["name"] == flat_state_path:
            return input
    else:
        return None


def validate_test_cases_for_tool_source(
    tool_source: ToolSource, use_latest_profile: bool = False, name: Optional[str] = None
) -> List[TestCaseStateValidationResult]:
    name = name or f"PydanticModelFor[{tool_source.parse_id()}]"
    tool_parameter_bundle = input_models_for_tool_source(tool_source)
    if use_latest_profile:
        # this might get old but it is fine, just needs to be updated when test case changes are made
        profile = "24.2"
    else:
        profile = tool_source.parse_profile()
    test_cases: List[ToolSourceTest] = tool_source.parse_tests_to_dict()["tests"]
    results_by_test: List[TestCaseStateValidationResult] = []
    for test_case in test_cases:
        validation_result = test_case_validation(test_case, tool_parameter_bundle.parameters, profile, name=name)
        results_by_test.append(validation_result)
    return results_by_test



"""This module contains linters for a tool's command description.

A command description describes how to build the command-line to execute
from supplied inputs.
"""

from typing import TYPE_CHECKING

from galaxy.tool_util.lint import Linter

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource


class CommandMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        root = tool_xml.find("./command")
        if root is None:
            root = tool_xml.getroot()
        command = tool_xml.find("./command")
        if command is None:
            lint_ctx.error(
                "No command tag found, must specify a command template to execute.", linter=cls.name(), node=root
            )


class CommandEmpty(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        root = tool_xml.find("./command")
        if root is None:
            root = tool_xml.getroot()
        command = tool_xml.find("./command")
        if command is not None and command.text is None:
            lint_ctx.error("Command is empty.", linter=cls.name(), node=root)


class CommandTODO(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        command = tool_xml.find("./command")
        if command is not None and command.text is not None and "TODO" in command.text:
            lint_ctx.warn("Command template contains TODO text.", linter=cls.name(), node=command)


class CommandInterpreterDeprecated(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        command = tool_xml.find("./command")
        if command is None:
            return
        interpreter_type = command.attrib.get("interpreter", None)
        if interpreter_type is not None:
            lint_ctx.warn("Command uses deprecated 'interpreter' attribute.", linter=cls.name(), node=command)


class CommandInfo(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        command = tool_xml.find("./command")
        if command is None:
            return
        interpreter_type = command.attrib.get("interpreter", None)
        interpreter_info = ""
        if interpreter_type:
            interpreter_info = f" with interpreter of type [{interpreter_type}]"
        lint_ctx.info(f"Tool contains a command{interpreter_info}.", linter=cls.name(), node=command)



"""Module of utilities for verifying test results."""

import difflib
import filecmp
import hashlib
import json
import logging
import math
import os
import os.path
import re
import shutil
import tempfile
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

try:
    import numpy
except ImportError:
    pass
try:
    import pysam
except ImportError:
    pass
try:
    from PIL import Image
except ImportError:
    Image = None  # type: ignore[assignment, unused-ignore]
try:
    import tifffile
except ImportError:
    tifffile = None  # type: ignore[assignment, unused-ignore]


from galaxy.tool_util.parser.util import (
    DEFAULT_DELTA,
    DEFAULT_DELTA_FRAC,
    DEFAULT_EPS,
    DEFAULT_METRIC,
    DEFAULT_PIN_LABELS,
)
from galaxy.tool_util.parser.yaml import to_test_assert_list
from galaxy.util import unicodify
from galaxy.util.compression_utils import get_fileobj
from ._types import (
    ExpandedToolInputsJsonified,
    ToolTestDescriptionDict,
)
from .asserts import verify_assertions
from .test_data import TestDataResolver

if TYPE_CHECKING:
    import numpy.typing

log = logging.getLogger(__name__)

DEFAULT_TEST_DATA_RESOLVER = TestDataResolver()
GetFilenameT = Optional[Callable[[str], str]]
GetLocationT = Optional[Callable[[str], str]]


def verify(
    item_label: str,
    output_content: bytes,
    attributes: Optional[Dict[str, Any]],
    filename: Optional[str] = None,
    get_filecontent: Optional[Callable[[str], bytes]] = None,
    get_filename: GetFilenameT = None,
    keep_outputs_dir: Optional[str] = None,
    verify_extra_files: Optional[Callable] = None,
    mode="file",
):
    """Verify the content of a test output using test definitions described by attributes.

    Throw an informative assertion error if any of these tests fail.
    """
    attributes = attributes or {}
    if get_filename is None:
        get_filecontent_: Callable[[str], bytes]
        if get_filecontent is None:
            get_filecontent_ = DEFAULT_TEST_DATA_RESOLVER.get_filecontent
        else:
            get_filecontent_ = get_filecontent

        def get_filename(filename: str) -> str:
            file_content = get_filecontent_(filename)
            local_name = make_temp_fname(fname=filename)
            with open(local_name, "wb") as f:
                f.write(file_content)
            return local_name

    # Check assertions...
    assertions = attributes.get("assert_list", None)
    if assertions is not None:
        try:
            verify_assertions(output_content, attributes["assert_list"], attributes.get("decompress", False))
        except AssertionError as err:
            errmsg = f"{item_label} different than expected\n"
            errmsg += unicodify(err)
            raise AssertionError(errmsg)

    # Verify checksum attributes...
    # works with older Galaxy style md5=<expected_sum> or cwltest
    # style checksum=<hash_type>$<hash>.
    expected_checksum_type = None
    expected_checksum = None
    if attributes is not None and attributes.get("md5", None) is not None:
        expected_checksum_type = "md5"
        expected_checksum = attributes.get("md5")
    elif attributes is not None and attributes.get("checksum") is not None:
        checksum_value = attributes.get("checksum")
        assert (
            checksum_value is not None
        )  # redundant with the test in the elif, but cannot use := until we drop support for Python 3.7
        expected_checksum_type, expected_checksum = checksum_value.split("$", 1)

    if expected_checksum_type:
        try:
            _verify_checksum(output_content, expected_checksum_type, expected_checksum)
        except AssertionError as err:
            errmsg = f"{item_label} different than expected\n"
            errmsg += unicodify(err)
            raise AssertionError(errmsg)

    # expected object might be None, so don't pull unless available
    has_expected_object = "object" in attributes
    if has_expected_object:
        assert filename is None
        expected_object = attributes.get("object")
        actual_object = json.loads(output_content)

        expected_object_type = type(expected_object)
        actual_object_type = type(actual_object)

        if expected_object_type != actual_object_type:
            message = f"Type mismatch between expected object ({expected_object_type}) and actual object ({actual_object_type})"
            raise AssertionError(message)

        if expected_object != actual_object:
            message = f"Expected object ({expected_object}) does not match actual object ({actual_object})"
            raise AssertionError(message)

    elif filename is not None:
        temp_name = make_temp_fname(fname=filename)
        with open(temp_name, "wb") as f:
            f.write(output_content)

        # If the server's env has GALAXY_TEST_SAVE, save the output file to that
        # directory.
        # This needs to be done before the call to `get_filename()` because that
        # may raise an exception if `filename` does not exist (e.g. when
        # generating a tool output file from scratch with
        # `planemo test --update_test_data`).
        if keep_outputs_dir:
            ofn = os.path.join(keep_outputs_dir, filename)
            out_dir = os.path.dirname(ofn)
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            log.debug("keep_outputs_dir: %s, ofn: %s", keep_outputs_dir, ofn)
            try:
                shutil.copy(temp_name, ofn)
            except Exception:
                log.exception("Could not save output file %s to %s", temp_name, ofn)
            else:
                log.debug("## GALAXY_TEST_SAVE=%s. saved %s", keep_outputs_dir, ofn)

        if mode == "directory":
            # if verifying a file inside a extra_files_path directory
            # filename already point to a file that exists on disk
            local_name = filename
        else:
            filename_ = get_filename(filename)
            assert filename_, f"Failed to find output target for test {filename_}"
            local_name = filename_

        compare = attributes.get("compare", "diff")
        try:
            if attributes.get("ftype", None) in [
                "bam",
                "qname_sorted.bam",
                "qname_input_sorted.bam",
                "unsorted.bam",
                "cram",
            ]:
                try:
                    local_fh, temp_name = _bam_to_sam(local_name, temp_name)
                    local_name = local_fh.name
                except Exception as e:
                    log.warning("%s. Will compare BAM files", unicodify(e))
            if compare == "diff":
                files_diff(local_name, temp_name, attributes=attributes)
            elif compare == "re_match":
                files_re_match(local_name, temp_name, attributes=attributes)
            elif compare == "re_match_multiline":
                files_re_match_multiline(local_name, temp_name, attributes=attributes)
            elif compare == "sim_size":
                files_delta(local_name, temp_name, attributes=attributes)
            elif compare == "contains":
                files_contains(local_name, temp_name, attributes=attributes)
            elif compare == "image_diff":
                if Image and tifffile:
                    files_image_diff(local_name, temp_name, attributes=attributes)
                else:
                    raise Exception(
                        "pillow and tifffile are not installed, but required to compare files using the 'image_diff' method"
                    )
            else:
                raise Exception(f"Unimplemented Compare type: {compare}")
        except AssertionError as err:
            errmsg = f"{item_label} different than expected, difference (using {compare}):\n"
            errmsg += f"( {local_name} v. {temp_name} )\n"
            errmsg += unicodify(err)
            raise AssertionError(errmsg)
        finally:
            if "GALAXY_TEST_NO_CLEANUP" not in os.environ:
                os.remove(temp_name)

    if verify_extra_files:
        extra_files = attributes.get("extra_files", None)
        if extra_files:
            verify_extra_files(extra_files)


def make_temp_fname(fname=None):
    """Safe temp name - preserve the file extension for tools that interpret it."""
    suffix = os.path.split(fname)[-1]  # ignore full path
    with tempfile.NamedTemporaryFile(prefix="tmp", suffix=suffix, delete=False) as temp:
        return temp.name


def _bam_to_sam(local_name, temp_name):
    temp_local = tempfile.NamedTemporaryFile(suffix=".sam", prefix="local_bam_converted_to_sam_")
    with tempfile.NamedTemporaryFile(suffix=".sam", prefix="history_bam_converted_to_sam_", delete=False) as temp:
        try:
            pysam.view("-h", "--no-PG", "-o", temp_local.name, local_name, catch_stdout=False)
        except Exception as e:
            msg = f"Converting local (test-data) BAM to SAM failed: {unicodify(e)}"
            raise Exception(msg)
        try:
            pysam.view("-h", "--no-PG", "-o", temp.name, temp_name, catch_stdout=False)
        except Exception as e:
            msg = f"Converting history BAM to SAM failed: {unicodify(e)}"
            raise Exception(msg)
    os.remove(temp_name)
    return temp_local, temp.name


def _verify_checksum(data, checksum_type, expected_checksum_value):
    if checksum_type not in ["md5", "sha1", "sha256", "sha512"]:
        raise Exception(f"Unimplemented hash algorithm [{checksum_type}] encountered.")

    h = hashlib.new(checksum_type)
    h.update(data)
    actual_checksum_value = h.hexdigest()
    if expected_checksum_value != actual_checksum_value:
        template = "Output checksum [%s] does not match expected [%s] (using hash algorithm %s)."
        message = template % (actual_checksum_value, expected_checksum_value, checksum_type)
        raise AssertionError(message)


def files_delta(file1, file2, attributes=None):
    """Check the contents of 2 files for size differences."""
    if attributes is None:
        attributes = {}
    delta = attributes.get("delta", DEFAULT_DELTA)
    delta_frac = attributes.get("delta_frac", DEFAULT_DELTA_FRAC)
    s1 = os.path.getsize(file1)
    s2 = os.path.getsize(file2)
    if abs(s1 - s2) > delta:
        raise AssertionError(f"Files {file1}={s1}b but {file2}={s2}b - compare by size (delta={delta}) failed")
    if delta_frac is not None and not (s1 - (s1 * delta_frac) <= s2 <= s1 + (s1 * delta_frac)):
        raise AssertionError(
            f"Files {file1}={s1}b but {file2}={s2}b - compare by size (delta_frac={delta_frac}) failed"
        )


def get_compressed_formats(attributes):
    attributes = attributes or {}
    decompress = attributes.get("decompress")
    # None means all compressed formats are allowed
    return None if decompress else []


def files_diff(file1: str, file2: str, attributes=None):
    """Check the contents of 2 files for differences."""
    attributes = attributes or {}

    def get_lines_diff(diff):
        count = 0
        for line in diff:
            if (line.startswith("+") and not line.startswith("+++")) or (
                line.startswith("-") and not line.startswith("---")
            ):
                count += 1
        return count

    if not filecmp.cmp(file1, file2, shallow=False):
        compressed_formats = get_compressed_formats(attributes)
        is_pdf = False
        try:
            with get_fileobj(file2, compressed_formats=compressed_formats, mode="r") as fh:
                history_data = fh.readlines()
            with get_fileobj(file1, compressed_formats=compressed_formats, mode="r") as fh:
                local_file = fh.readlines()
        except UnicodeDecodeError:
            if file1.endswith(".pdf") or file2.endswith(".pdf"):
                is_pdf = True
                # Replace non-Unicode characters using unicodify(),
                # difflib.unified_diff doesn't work on list of bytes
                history_data = [
                    unicodify(line) for line in get_fileobj(file2, mode="rb", compressed_formats=compressed_formats)
                ]
                local_file = [
                    unicodify(line) for line in get_fileobj(file1, mode="rb", compressed_formats=compressed_formats)
                ]
            else:
                raise AssertionError("Binary data detected, not displaying diff")
        if attributes.get("sort", False):
            local_file.sort()
            history_data.sort()
        allowed_diff_count = int(attributes.get("lines_diff", 0))
        diff = list(difflib.unified_diff(local_file, history_data, "local_file", "history_data"))
        diff_lines = get_lines_diff(diff)
        if diff_lines > allowed_diff_count:
            if "GALAXY_TEST_RAW_DIFF" in os.environ:
                diff_slice = diff
            else:
                if len(diff) < 60:
                    diff_slice = diff[0:40]
                else:
                    diff_slice = diff[:25] + ["********\n", "*SNIP *\n", "********\n"] + diff[-25:]
            # FIXME: This pdf stuff is rather special cased and has not been updated to consider lines_diff
            # due to unknown desired behavior when used in conjunction with a non-zero lines_diff
            # PDF forgiveness can probably be handled better by not special casing by __extension__ here
            # and instead using lines_diff or a regular expression matching
            # or by creating and using a specialized pdf comparison function
            if is_pdf:
                # PDF files contain creation dates, modification dates, ids and descriptions that change with each
                # new file, so we need to handle these differences.  As long as the rest of the PDF file does
                # not differ we're ok.
                valid_diff_strs = ["description", "createdate", "creationdate", "moddate", "id", "producer", "creator"]
                valid_diff = False
                invalid_diff_lines = 0
                for line in diff_slice:
                    # Make sure to lower case strings before checking.
                    line = line.lower()
                    # Diff lines will always start with a + or - character, but handle special cases: '--- local_file \n', '+++ history_data \n'
                    if (
                        (line.startswith("+") or line.startswith("-"))
                        and line.find("local_file") < 0
                        and line.find("history_data") < 0
                    ):
                        for vdf in valid_diff_strs:
                            if line.find(vdf) < 0:
                                valid_diff = False
                            else:
                                valid_diff = True
                                # Stop checking as soon as we know we have a valid difference
                                break
                        if not valid_diff:
                            invalid_diff_lines += 1
                log.info(
                    "## files diff on '%s' and '%s': lines_diff = %d, found diff = %d, found pdf invalid diff = %d",
                    file1,
                    file2,
                    allowed_diff_count,
                    diff_lines,
                    invalid_diff_lines,
                )
                if invalid_diff_lines > allowed_diff_count:
                    # Print out diff_slice so we can see what failed
                    log.info("###### diff_slice ######")
                    raise AssertionError("".join(diff_slice))
            else:
                log.info(
                    "## files diff on '%s' and '%s': lines_diff = %d, found diff = %d",
                    file1,
                    file2,
                    allowed_diff_count,
                    diff_lines,
                )
                raise AssertionError("".join(diff_slice))


def files_re_match(file1, file2, attributes=None):
    """Check the contents of 2 files for differences using re.match."""
    attributes = attributes or {}
    join_char = ""
    to_strip = os.linesep
    compressed_formats = get_compressed_formats(attributes)
    try:
        with get_fileobj(file2, compressed_formats=compressed_formats) as fh:
            history_data = fh.readlines()
        with get_fileobj(file1, compressed_formats=compressed_formats) as fh:
            local_file = fh.readlines()
    except UnicodeDecodeError:
        join_char = b""
        to_strip = os.linesep.encode("utf-8")
        with open(file2, "rb") as fh:
            history_data = fh.readlines()
        with open(file1, "rb") as fh:
            local_file = fh.readlines()
    assert len(local_file) == len(
        history_data
    ), f"Data File and Regular Expression File contain a different number of lines ({len(local_file)} != {len(history_data)})\nHistory Data (first 40 lines):\n{join_char.join(history_data[:40])}"
    if attributes.get("sort", False):
        history_data.sort()
        local_file.sort()
    lines_diff = int(attributes.get("lines_diff", 0))
    line_diff_count = 0
    diffs = []
    for regex_line, data_line in zip(local_file, history_data):
        regex_line = regex_line.rstrip(to_strip)
        data_line = data_line.rstrip(to_strip)
        if not re.match(regex_line, data_line):
            line_diff_count += 1
            diffs.append(f"Regular Expression: {regex_line}, Data file: {data_line}\n")
    if line_diff_count > lines_diff:
        raise AssertionError(
            "Regular expression did not match data file (allowed variants={}):\n{}".format(lines_diff, "".join(diffs))
        )


def files_re_match_multiline(file1, file2, attributes=None):
    """Check the contents of 2 files for differences using re.match in multiline mode."""
    attributes = attributes or {}
    join_char = ""
    compressed_formats = get_compressed_formats(attributes)
    try:
        with get_fileobj(file2, compressed_formats=compressed_formats) as fh:
            history_data = fh.readlines()
        with get_fileobj(file1, compressed_formats=compressed_formats) as fh:
            local_file = fh.read()
    except UnicodeDecodeError:
        join_char = b""
        with open(file2, "rb") as fh:
            history_data = fh.readlines()
        with open(file1, "rb") as fh:
            local_file = fh.read()
    if attributes.get("sort", False):
        history_data.sort()
    history_data = join_char.join(history_data)
    # lines_diff not applicable to multiline matching
    assert re.match(local_file, history_data, re.MULTILINE), "Multiline Regular expression did not match data file"


def files_contains(file1, file2, attributes=None):
    """Check the contents of file2 for substrings found in file1, on a per-line basis."""
    # TODO: allow forcing ordering of contains
    attributes = attributes or {}
    to_strip = os.linesep
    compressed_formats = get_compressed_formats(attributes)
    try:
        with get_fileobj(file2, compressed_formats=compressed_formats) as fh:
            history_data = fh.read()
        with get_fileobj(file1, compressed_formats=compressed_formats) as fh:
            local_file = fh.readlines()
    except UnicodeDecodeError:
        to_strip = os.linesep.encode("utf-8")
        with open(file2, "rb") as fh:
            history_data = fh.read()
        with open(file1, "rb") as fh:
            local_file = fh.readlines()
    lines_diff = int(attributes.get("lines_diff", 0))
    line_diff_count = 0
    for contains in local_file:
        contains = contains.rstrip(to_strip)
        if contains not in history_data:
            line_diff_count += 1
        if line_diff_count > lines_diff:
            raise AssertionError(f"Failed to find '{contains}' in history data. (lines_diff={lines_diff}).")


def _singleobject_intersection_over_union(
    mask1: "numpy.typing.NDArray[numpy.bool_]",
    mask2: "numpy.typing.NDArray[numpy.bool_]",
) -> "numpy.floating":
    return numpy.logical_and(mask1, mask2).sum() / numpy.logical_or(mask1, mask2).sum()


def _multiobject_intersection_over_union(
    mask1: "numpy.typing.NDArray",
    mask2: "numpy.typing.NDArray",
    pin_labels: Optional[List[int]] = None,
    repeat_reverse: bool = True,
) -> List["numpy.floating"]:
    iou_list: List[numpy.floating] = []
    for label1 in numpy.unique(mask1):
        cc1 = mask1 == label1

        # If the label is in `pin_labels`, then use the same label value to find the corresponding object in the second mask.
        if pin_labels is not None and label1 in pin_labels:
            cc2 = mask2 == label1
            iou_list.append(_singleobject_intersection_over_union(cc1, cc2))

        # Otherwise, use the object with the largest IoU value, excluding the pinned labels.
        else:
            cc1_iou_list: List[numpy.floating] = []
            for label2 in numpy.unique(mask2[cc1]):
                if pin_labels is not None and label2 in pin_labels:
                    continue
                cc2 = mask2 == label2
                cc1_iou_list.append(_singleobject_intersection_over_union(cc1, cc2))
            iou_list.append(max(cc1_iou_list))  # type: ignore[type-var, unused-ignore]  # https://github.com/python/typeshed/issues/12562

    if repeat_reverse:
        iou_list.extend(_multiobject_intersection_over_union(mask2, mask1, pin_labels, repeat_reverse=False))

    return iou_list


def intersection_over_union(
    mask1: "numpy.typing.NDArray", mask2: "numpy.typing.NDArray", pin_labels: Optional[List[int]] = None
) -> "numpy.floating":
    """Compute the intersection over union (IoU) for the objects in two masks containing labels.

    The IoU is computed for each uniquely labeled image region (object), and the overall minimum value is returned (i.e. the worst value).
    To compute the IoU for each object, the corresponding object in the other mask needs to be determined.
    The object correspondences are not necessarily symmetric.

    By default, the corresponding object in the other mask is determined as the one with the largest IoU value.
    If the label of an object is listed in `pin_labels`, then the corresponding object in the other mask is determined as the object with the same label value.
    Objects with labels listed in `pin_labels` also cannot correspond to objects with different labels.
    This is particularly useful when specific image regions must always be labeled with a designated label value (e.g., the image background is often labeled with 0 or -1).
    """
    assert mask1.dtype == mask2.dtype
    assert mask1.ndim == mask2.ndim == 2
    assert mask1.shape == mask2.shape
    for label in pin_labels or []:
        count = sum(label in mask for mask in (mask1, mask2))
        count_str = {1: "one", 2: "both"}
        assert count == 2, f"Label {label} is pinned but missing in {count_str[2 - count]} of the images."
    return min(_multiobject_intersection_over_union(mask1, mask2, pin_labels))  # type: ignore[type-var, unused-ignore]  # https://github.com/python/typeshed/issues/12562


def _parse_label_list(label_list_str: Optional[str]) -> List[int]:
    if label_list_str is None:
        return []
    else:
        return [int(label.strip()) for label in label_list_str.split(",") if len(label_list_str) > 0]


def get_image_metric(
    attributes: Dict[str, Any],
) -> Callable[["numpy.typing.NDArray", "numpy.typing.NDArray"], "numpy.floating"]:
    metric_name = attributes.get("metric", DEFAULT_METRIC)
    pin_labels = _parse_label_list(attributes.get("pin_labels", DEFAULT_PIN_LABELS))
    metrics = {
        "mae": lambda arr1, arr2: numpy.abs(arr1 - arr2).mean(),
        # Convert to float before squaring to prevent overflows
        "mse": lambda arr1, arr2: numpy.square((arr1 - arr2).astype(float)).mean(),
        "rms": lambda arr1, arr2: math.sqrt(numpy.square((arr1 - arr2).astype(float)).mean()),
        "fro": lambda arr1, arr2: numpy.linalg.norm((arr1 - arr2).reshape(1, -1), "fro"),
        "iou": lambda arr1, arr2: 1 - intersection_over_union(arr1, arr2, pin_labels),
    }
    try:
        return metrics[metric_name]
    except KeyError:
        raise ValueError(f'No such metric: "{metric_name}"')


def _load_image(filepath: str) -> "numpy.typing.NDArray":
    """
    Reads the given image, trying tifffile and Pillow for reading.
    """
    # Try reading with tifffile first. It fails if the file is not a TIFF.
    try:
        arr = tifffile.imread(filepath)

    # If tifffile failed, then the file is not a tifffile. In that case, try with Pillow.
    except tifffile.TiffFileError:
        with Image.open(filepath) as im:
            arr = numpy.array(im)

    # Return loaded image
    return arr


def files_image_diff(file1: str, file2: str, attributes: Optional[Dict[str, Any]] = None) -> None:
    """Check the pixel data of 2 image files for differences."""
    attributes = attributes or {}

    arr1 = _load_image(file1)
    arr2 = _load_image(file2)

    if arr1.dtype != arr2.dtype:
        raise AssertionError(f"Image data types did not match ({arr1.dtype}, {arr2.dtype}).")

    if arr1.shape != arr2.shape:
        raise AssertionError(f"Image dimensions did not match ({arr1.shape}, {arr2.shape}).")

    # Handle `bool` images by converting them to `uint8`
    if numpy.issubdtype(arr1.dtype, bool):
        arr1 = arr1.astype(numpy.uint8)
    if numpy.issubdtype(arr2.dtype, bool):
        arr2 = arr2.astype(numpy.uint8)

    distance = get_image_metric(attributes)(arr1, arr2)
    distance_eps = attributes.get("eps", DEFAULT_EPS)
    if distance > distance_eps:
        raise AssertionError(f"Image difference {distance} exceeds eps={distance_eps}.")


# TODO: After tool-util with this included is published, fefactor planemo.test._check_output
# to use this function. There is already a comment there about breaking fewer abstractions.
# https://github.com/galaxyproject/planemo/blob/master/planemo/test/_check_output.py
# TODO: Also migrate the logic for checking non-dictionaries out of Planemo - this function now
# does that check also.
def verify_file_path_against_dict(
    get_filename: GetFilenameT,
    get_location: GetLocationT,
    path: str,
    output_content: bytes,
    test_properties,
    test_data_target_dir: Optional[str] = None,
) -> None:
    with open(path, "rb") as f:
        output_content = f.read()
    item_label = f"Output with path {path}"
    verify_file_contents_against_dict(
        get_filename, get_location, item_label, output_content, test_properties, test_data_target_dir
    )


def verify_file_contents_against_dict(
    get_filename: GetFilenameT,
    get_location: GetLocationT,
    item_label: str,
    output_content: bytes,
    test_properties,
    test_data_target_dir: Optional[str] = None,
) -> None:
    expected_file: Optional[str] = None
    if isinstance(test_properties, dict):
        # Support Galaxy-like file location (using "file") or CWL-like ("path" or "location").
        expected_file = test_properties.get("file", None)
        if expected_file is None:
            expected_file = test_properties.get("path", None)
        if expected_file is None:
            location = test_properties.get("location")
            if location:
                if location.startswith(("http://", "https://")):
                    assert get_location
                    expected_file = get_location(location)
                else:
                    expected_file = location.split("file://", 1)[-1]

        if "asserts" in test_properties:
            test_properties["assert_list"] = to_test_assert_list(test_properties["asserts"])
        verify(
            item_label,
            output_content,
            attributes=test_properties,
            filename=expected_file,
            get_filename=get_filename,
            keep_outputs_dir=test_data_target_dir,
            verify_extra_files=None,
        )
    else:
        output_value = json.loads(output_content.decode("utf-8"))
        if test_properties != output_value:
            template = "Output [%s] value [%s] does not match expected value [%s]."
            message = template % (item_label, output_value, test_properties)
            raise AssertionError(message)


__all__ = [
    "DEFAULT_TEST_DATA_RESOLVER",
    "ExpandedToolInputsJsonified",
    "GetFilenameT",
    "GetLocationT",
    "ToolTestDescriptionDict",
    "verify",
    "verify_file_contents_against_dict",
]



"""This module contains a linting functions for tool outputs."""

import ast
from typing import TYPE_CHECKING

from packaging.version import Version

from galaxy.tool_util.lint import Linter
from ._util import is_valid_cheetah_placeholder
from ..parser.output_collection_def import NAMED_PATTERNS

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser import ToolSource
    from galaxy.util.etree import (
        Element,
        ElementTree,
    )


class OutputsMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_node = tool_xml.find("./outputs")
        if tool_node is None:
            tool_node = tool_xml.getroot()
        if len(tool_xml.findall("./outputs")) == 0:
            lint_ctx.warn(
                "Tool contains no outputs section, most tools should produce outputs.",
                linter=cls.name(),
                node=tool_node,
            )


class OutputsOutput(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        output = tool_xml.find("./outputs/output")
        if output is not None:
            lint_ctx.warn(
                "Avoid the use of 'output' and replace by 'data' or 'collection'", linter=cls.name(), node=output
            )


class OutputsNameInvalidCheetah(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for output in tool_xml.findall("./outputs/data[@name]") + tool_xml.findall("./outputs/collection[@name]"):
            if not is_valid_cheetah_placeholder(output.attrib["name"]):
                lint_ctx.warn(
                    f'Tool output name [{output.attrib["name"]}] is not a valid Cheetah placeholder.',
                    linter=cls.name(),
                    node=output,
                )


class OutputsNameDuplicated(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        names = set()
        for output in tool_xml.findall("./outputs/data[@name]") + tool_xml.findall("./outputs/collection[@name]"):
            name = output.attrib["name"]
            if name in names:
                lint_ctx.error(f"Tool output [{name}] has duplicated name", linter=cls.name(), node=output)
            names.add(name)


class OutputsFilterExpression(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for filter in tool_xml.findall("./outputs/*/filter"):
            try:
                ast.parse(filter.text.strip(), mode="eval")
            except Exception as e:
                lint_ctx.warn(
                    f"Filter '{filter.text}' is no valid expression: {str(e)}",
                    linter=cls.name(),
                    node=filter,
                )


class OutputsLabelDuplicatedFilter(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        labels = set()
        for output in tool_xml.findall("./outputs/data") + tool_xml.findall("./outputs/collection"):
            name = output.attrib.get("name", "")
            label = output.attrib.get("label", "${tool.name} on ${on_string}")
            if label in labels and output.find("./filter") is not None:
                lint_ctx.warn(
                    f"Tool output [{name}] uses duplicated label '{label}', double check if filters imply disjoint cases",
                    linter=cls.name(),
                    node=output,
                )
            labels.add(label)


class OutputsLabelDuplicatedNoFilter(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        labels = set()
        for output in tool_xml.findall("./outputs/data[@name]") + tool_xml.findall("./outputs/collection[@name]"):
            name = output.attrib.get("name", "")
            label = output.attrib.get("label", "${tool.name} on ${on_string}")
            if label in labels and output.find("./filter") is None:
                lint_ctx.warn(f"Tool output [{name}] uses duplicated label '{label}'", linter=cls.name(), node=output)
            labels.add(label)


class OutputsCollectionType(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for output in tool_xml.findall("./outputs/collection"):
            if "type" not in output.attrib:
                lint_ctx.warn("Collection output with undefined 'type' found.", linter=cls.name(), node=output)


class OutputsNumber(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        outputs = tool_xml.findall("./outputs")
        if len(outputs) == 0:
            return
        num_outputs = len(outputs[0].findall("./data")) + len(outputs[0].findall("./collection"))
        lint_ctx.info(f"{num_outputs} outputs found.", linter=cls.name(), node=outputs[0])


class OutputsFormatInput(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        def _report(output: "Element"):
            message = f"Using format='input' on {output.tag} is deprecated. Use the format_source attribute."
            if Version(str(profile)) <= Version("16.01"):
                lint_ctx.warn(message, linter=cls.name(), node=output)
            else:
                lint_ctx.error(message, linter=cls.name(), node=output)

        profile = tool_source.parse_profile()
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for output in tool_xml.findall("./outputs/data") + tool_xml.findall("./outputs/collection"):
            fmt = output.attrib.get("format")
            if fmt == "input":
                _report(output)
            for sub in output:
                fmt = sub.attrib.get("format", sub.attrib.get("ext"))
                if fmt == "input":
                    _report(output)


class OutputsFormat(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for output in tool_xml.findall("./outputs/data") + tool_xml.findall("./outputs/collection"):
            format_set = False
            if _check_format(output):
                format_set = True
            if output.tag == "data":
                if "auto_format" in output.attrib and output.attrib["auto_format"]:
                    format_set = True

            elif output.tag == "collection":
                if "structured_like" in output.attrib and "inherit_format" in output.attrib:
                    format_set = True
            for sub in output:
                if _check_pattern(sub) or _has_tool_provided_metadata(tool_xml):
                    format_set = True
                elif _check_format(sub):
                    format_set = True

            if not format_set:
                lint_ctx.warn(
                    f"Tool {output.tag} output {output.attrib.get('name', 'with missing name')} doesn't define an output format.",
                    linter=cls.name(),
                    node=output,
                )


class OutputsFormatSourceIncomp(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        def _check_and_report(node):
            if "format_source" in node.attrib and ("ext" in node.attrib or "format" in node.attrib):
                lint_ctx.warn(
                    f"Tool {node.tag} output '{node.attrib.get('name', 'with missing name')}' should use either format_source or format/ext",
                    linter=cls.name(),
                    node=node,
                )

        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for output in tool_xml.findall("./outputs/data") + tool_xml.findall("./outputs/collection"):
            _check_and_report(output)
            for sub in output:
                _check_and_report(sub)


def _check_format(node):
    """
    check if format/ext/format_source attribute is set in a given node
    issue a warning if the value is input
    return true (node defines format/ext) / false (else)
    """
    if "format_source" in node.attrib:
        return True
    if node.find(".//action[@type='format']") is not None:
        return True
    # if allowed (e.g. for discover_datasets), ext takes precedence over format
    fmt = node.attrib.get("format", node.attrib.get("ext"))
    return fmt is not None


def _check_pattern(node):
    """
    check if
    - pattern attribute is set and defines the extension or
    - from_tool_provided_metadata is true
    """
    if node.tag != "discover_datasets":
        return False
    if "pattern" not in node.attrib:
        return False
    pattern = node.attrib["pattern"]
    regex_pattern = NAMED_PATTERNS.get(pattern, pattern)
    # TODO error on wrong pattern or non-regexp
    if "(?P<ext>" in regex_pattern:
        return True


class OutputsStructuredLikeReference(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        param_qualified_paths = _collect_param_qualified_paths(tool_xml)
        for output in tool_xml.findall("./outputs/collection[@structured_like]"):
            structured_like = output.attrib["structured_like"]
            _check_unqualified_reference(
                lint_ctx, cls.name(), output, structured_like, "structured_like", param_qualified_paths
            )


class OutputsFormatSourceReference(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        param_qualified_paths = _collect_param_qualified_paths(tool_xml)
        output_names = {
            o.attrib["name"]
            for o in tool_xml.findall("./outputs/data[@name]") + tool_xml.findall("./outputs/collection[@name]")
        }
        for output in tool_xml.findall("./outputs/data[@format_source]") + tool_xml.findall(
            "./outputs/collection[@format_source]"
        ):
            format_source = output.attrib["format_source"]
            # format_source can reference other outputs, skip if it matches an output name
            if format_source in output_names:
                continue
            _check_unqualified_reference(
                lint_ctx, cls.name(), output, format_source, "format_source", param_qualified_paths
            )


def _check_unqualified_reference(
    lint_ctx: "LintContext",
    linter_name: str,
    node: "Element",
    ref_value: str,
    attr_name: str,
    param_qualified_paths: dict,
):
    if "|" in ref_value:
        return
    # Check if it matches a top-level param directly
    top_level_match = any(qp == ref_value for paths in param_qualified_paths.values() for qp in paths)
    if top_level_match:
        return
    matches = param_qualified_paths.get(ref_value, [])
    output_name = node.attrib.get("name", "unknown")
    if len(matches) == 1:
        lint_ctx.warn(
            f"Output '{output_name}' uses unqualified {attr_name}='{ref_value}'. "
            f"Use the qualified name '{matches[0]}'.",
            linter=linter_name,
            node=node,
        )
    elif len(matches) > 1:
        lint_ctx.warn(
            f"Output '{output_name}' uses ambiguous unqualified {attr_name}='{ref_value}' "
            f"matching multiple inputs: {', '.join(matches)}. Use a qualified name.",
            linter=linter_name,
            node=node,
        )
    else:
        lint_ctx.error(
            f"Output '{output_name}' references {attr_name}='{ref_value}' "
            f"which does not match any input parameter.",
            linter=linter_name,
            node=node,
        )


def _collect_param_qualified_paths(tool_xml: "ElementTree") -> dict:
    """Build a map of unqualified param name -> list of qualified paths."""
    param_paths: dict = {}
    parent_map = {child: parent for parent in tool_xml.iter() for child in parent}
    for param in tool_xml.findall("./inputs//param"):
        name = param.attrib.get("name")
        if not name:
            argument = param.attrib.get("argument")
            if argument:
                name = argument.lstrip("-").replace("-", "_")
        if not name:
            continue
        qualified = _get_qualified_name(param, parent_map)
        param_paths.setdefault(name, []).append(qualified)
    return param_paths


def _get_qualified_name(param_elem: "Element", parent_map: dict) -> str:
    """Walk up the XML tree to build the qualified path for a param element."""
    name = param_elem.attrib.get("name")
    if not name:
        argument = param_elem.attrib.get("argument")
        if argument:
            name = argument.lstrip("-").replace("-", "_")
    parts = [name] if name else []
    current = param_elem
    while True:
        parent = parent_map.get(current)
        if parent is None:
            break
        if parent.tag in ("conditional", "section"):
            parent_name = parent.attrib.get("name")
            if parent_name:
                parts.insert(0, parent_name)
        elif parent.tag in ("inputs", "tool"):
            break
        current = parent
    return "|".join(parts)


def _has_tool_provided_metadata(tool_xml: "ElementTree") -> bool:
    outputs = tool_xml.find("./outputs")
    if outputs is not None:
        if "provided_metadata_file" in outputs.attrib or "provided_metadata_style" in outputs.attrib:
            return True
    command = tool_xml.find("./command")
    if command is not None:
        if "galaxy.json" in command.text:
            return True
    config = tool_xml.find("./configfiles/configfile[@filename='galaxy.json']")
    if config is not None:
        return True
    return False



import logging
import os
import traceback
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from packaging.version import Version

from galaxy.tool_util.parameters import (
    input_models_for_tool_source,
    test_case_state as case_state,
    TestCaseToolState,
    ToolParameterBundleModel,
)
from galaxy.tool_util.parser.interface import (
    InputSource,
    TestCollectionDef,
    ToolSource,
    ToolSourceTest,
    ToolSourceTestInputs,
    ToolSourceTests,
)
from galaxy.tool_util.parser.util import (
    boolean_is_checked,
    boolean_true_and_false_values,
    parse_tool_version_with_defaults,
)
from galaxy.tool_util.parser.xml import __parse_assert_list_from_elem
from galaxy.tool_util.verify.interactor import (
    InvalidToolTestDict,
    ToolTestDescription,
    ValidToolTestDict,
)
from galaxy.tool_util_models.assertions import relaxed_assertion_list
from galaxy.util import (
    string_as_bool,
    string_as_bool_or_none,
    unicodify,
)
from ._types import (
    ExpandedToolInputs,
    ExtraFileInfoDictT,
    RequiredDataTablesT,
    RequiredFilesT,
    RequiredLocFileT,
)

log = logging.getLogger(__name__)

AnyParamContext = Union["ParamContext", "RootParamContext"]


def parse_tool_test_descriptions(
    tool_source: ToolSource, tool_guid: Optional[str] = None
) -> Iterable[ToolTestDescription]:
    """
    Build ToolTestDescription objects for each test description.
    """
    validate_on_load = Version(tool_source.parse_profile()) >= Version("24.2")
    raw_tests_dict: ToolSourceTests = tool_source.parse_tests_to_dict()
    tests: List[ToolTestDescription] = []

    profile = tool_source.parse_profile()
    for i, raw_test_dict in enumerate(raw_tests_dict.get("tests", [])):
        validation_exception: Optional[Exception] = None
        request_and_schema: Optional[TestRequestAndSchema] = None
        if validate_on_load:
            try:
                tool_parameter_bundle = input_models_for_tool_source(tool_source)
                validated_test_case = case_state(
                    raw_test_dict, tool_parameter_bundle.parameters, profile, validate=True
                )
                request_and_schema = TestRequestAndSchema(
                    validated_test_case.tool_state,
                    tool_parameter_bundle,
                )
            except Exception as e:
                validation_exception = e

        if validation_exception:
            tool_id, tool_version = _tool_id_and_version(tool_source, tool_guid)
            test = ToolTestDescription.from_tool_source_dict(
                InvalidToolTestDict(
                    {
                        "tool_id": tool_id,
                        "tool_version": tool_version,
                        "test_index": i,
                        "inputs": {},
                        "error": True,
                        "exception": unicodify(validation_exception),
                        "maxseconds": None,
                    }
                )
            )
        else:
            test = _description_from_tool_source(tool_source, raw_test_dict, i, tool_guid, request_and_schema)
        tests.append(test)
    return tests


@dataclass
class TestRequestAndSchema:
    request: TestCaseToolState
    request_schema: ToolParameterBundleModel


def _description_from_tool_source(
    tool_source: ToolSource,
    raw_test_dict: ToolSourceTest,
    test_index: int,
    tool_guid: Optional[str],
    request_and_schema: Optional[TestRequestAndSchema],
) -> ToolTestDescription:
    required_files: RequiredFilesT = []
    required_data_tables: RequiredDataTablesT = []
    required_loc_files: RequiredLocFileT = []

    num_outputs = raw_test_dict.get("expect_num_outputs", None)
    if num_outputs:
        num_outputs = int(num_outputs)
    maxseconds = raw_test_dict.get("maxseconds", None)
    if maxseconds is not None:
        maxseconds = int(maxseconds)

    request: Optional[Dict[str, Any]] = None
    request_schema: Optional[Dict[str, Any]] = None
    if request_and_schema:
        request = request_and_schema.request.input_state
        request_schema = request_and_schema.request_schema.dict()

    tool_id, tool_version = _tool_id_and_version(tool_source, tool_guid)
    processed_test_dict: Union[ValidToolTestDict, InvalidToolTestDict]
    try:
        processed_inputs = _process_raw_inputs(
            tool_source,
            input_sources(tool_source),
            raw_test_dict["inputs"],
            required_files,
            required_data_tables,
            required_loc_files,
        )
        processed_test_dict = ValidToolTestDict(
            {
                "inputs": processed_inputs,
                "request": request,
                "request_schema": request_schema,
                "outputs": raw_test_dict["outputs"],
                "output_collections": raw_test_dict["output_collections"],
                "num_outputs": num_outputs,
                "command_line": raw_test_dict.get("command", None),
                "command_version": raw_test_dict.get("command_version", None),
                "stdout": raw_test_dict.get("stdout", None),
                "stderr": raw_test_dict.get("stderr", None),
                "expect_exit_code": raw_test_dict.get("expect_exit_code", None),
                "expect_failure": raw_test_dict.get("expect_failure", False),
                "expect_test_failure": raw_test_dict.get("expect_test_failure", False),
                "required_files": required_files,
                "required_data_tables": required_data_tables,
                "required_loc_files": required_loc_files,
                "tool_id": tool_id,
                "tool_version": tool_version,
                "test_index": test_index,
                "maxseconds": maxseconds,
                "error": False,
            }
        )
    except Exception:
        processed_test_dict = InvalidToolTestDict(
            {
                "tool_id": tool_id,
                "tool_version": tool_version,
                "test_index": test_index,
                "inputs": {},
                "error": True,
                "exception": unicodify(traceback.format_exc()),
                "maxseconds": maxseconds,
            }
        )

    return ToolTestDescription.from_tool_source_dict(processed_test_dict)


def _tool_id_and_version(tool_source: ToolSource, tool_guid: Optional[str]) -> Tuple[str, str]:
    tool_id = tool_guid or tool_source.parse_id()
    assert tool_id
    tool_version = parse_tool_version_with_defaults(tool_id, tool_source)
    return tool_id, tool_version


def _process_raw_inputs(
    tool_source: ToolSource,
    input_sources: List[InputSource],
    raw_inputs: ToolSourceTestInputs,
    required_files: RequiredFilesT,
    required_data_tables: RequiredDataTablesT,
    required_loc_files: RequiredLocFileT,
    parent_context: Optional[AnyParamContext] = None,
) -> ExpandedToolInputs:
    """
    Recursively expand flat list of inputs into "tree" form of flat list
    (| using to nest to new levels) structure and expand dataset
    information as proceeding to populate self.required_files.
    """
    profile = tool_source.parse_profile()
    allow_legacy_test_case_parameters = Version(profile) <= Version("24.1")
    parent_context = parent_context or RootParamContext(allow_unqualified_access=allow_legacy_test_case_parameters)
    expanded_inputs: ExpandedToolInputs = {}
    for input_source in input_sources:
        input_type = input_source.parse_input_type()
        name = input_source.parse_name()
        if input_type == "conditional":
            cond_context = ParamContext(name=name, parent_context=parent_context)
            test_param_input_source = input_source.parse_test_input_source()
            case_name = test_param_input_source.parse_name()
            case_context = ParamContext(name=case_name, parent_context=cond_context)
            raw_input_dict = case_context.extract_value(raw_inputs)
            case_value = raw_input_dict["value"] if raw_input_dict else None
            case_when, case_input_sources = _matching_case_for_value(
                tool_source, input_source, test_param_input_source, case_value, allow_legacy_test_case_parameters
            )
            if case_input_sources:
                for case_input_source in case_input_sources.parse_input_sources():
                    case_inputs = _process_raw_inputs(
                        tool_source,
                        [case_input_source],
                        raw_inputs,
                        required_files,
                        required_data_tables,
                        required_loc_files,
                        parent_context=cond_context,
                    )
                    expanded_inputs.update(case_inputs)
                expanded_case_value = split_if_str(case_when)
                if case_value is not None:
                    # A bit tricky here - we are growing inputs with value
                    # that may be implicit (i.e. not defined by user just
                    # a default defined in tool). So we do not want to grow
                    # expanded_inputs and risk repeat block viewing this
                    # as a new instance with value defined and hence enter
                    # an infinite loop - hence the "case_value is not None"
                    # check.
                    processed_value = _process_simple_value(
                        test_param_input_source,
                        expanded_case_value,
                        required_data_tables,
                        required_loc_files,
                        allow_legacy_test_case_parameters,
                    )
                    expanded_inputs[case_context.for_state()] = processed_value
        elif input_type == "section":
            context = ParamContext(name=name, parent_context=parent_context)
            page_source = input_source.parse_nested_inputs_source()
            for section_input_source in page_source.parse_input_sources():
                expanded_input = _process_raw_inputs(
                    tool_source,
                    [section_input_source],
                    raw_inputs,
                    required_files,
                    required_data_tables,
                    required_loc_files,
                    parent_context=context,
                )
                if expanded_input:
                    expanded_inputs.update(expanded_input)
        elif input_type == "repeat":
            repeat_index = 0
            while True:
                context = ParamContext(name=name, parent_context=parent_context, index=repeat_index)
                updated = False
                page_source = input_source.parse_nested_inputs_source()
                for r_value in page_source.parse_input_sources():
                    expanded_input = _process_raw_inputs(
                        tool_source,
                        [r_value],
                        raw_inputs,
                        required_files,
                        required_data_tables,
                        required_loc_files,
                        parent_context=context,
                    )
                    if expanded_input:
                        expanded_inputs.update(expanded_input)
                        updated = True
                if not updated:
                    break
                repeat_index += 1
        else:
            context = ParamContext(name=name, parent_context=parent_context)
            raw_input_dict = context.extract_value(raw_inputs)
            param_type = input_source.get("type")
            if raw_input_dict:
                name = raw_input_dict["name"]
                param_value = raw_input_dict["value"]
                param_extra = raw_input_dict["attributes"]
                location = param_extra.get("location")
                if param_type != "text":
                    param_value = split_if_str(param_value)
                if param_type in ("data", "hidden_data"):
                    if location and input_source.get_bool("multiple", False):
                        # We get the input/s from the location which can be a list of urls separated by commas
                        locations = split_if_str(location)
                        param_value = []
                        for location in locations:
                            v = os.path.basename(location)
                            param_value.append(v)
                            # param_extra should contain only the corresponding location
                            extra = dict(param_extra)
                            extra["location"] = location
                            _add_uploaded_dataset(context.for_state(), v, extra, input_source, required_files)
                    else:
                        if not isinstance(param_value, list):
                            param_value = [param_value]
                        for v in param_value:
                            _add_uploaded_dataset(context.for_state(), v, param_extra, input_source, required_files)
                    processed_value = param_value
                elif param_type == "data_collection":
                    assert "collection" in param_extra
                    collection_dict = param_extra["collection"]
                    collection_def = TestCollectionDef.from_dict(collection_dict)
                    for input_dict in collection_def.collect_inputs():
                        name = input_dict["name"]
                        value = input_dict["value"]
                        attributes = input_dict["attributes"]
                        require_file(name, value, attributes, required_files)
                    processed_value = collection_def
                else:
                    processed_value = _process_simple_value(
                        input_source,
                        param_value,
                        required_data_tables,
                        required_loc_files,
                        allow_legacy_test_case_parameters,
                    )
                expanded_inputs[context.for_state()] = processed_value
    return expanded_inputs


def input_sources(tool_source: ToolSource) -> List[InputSource]:
    input_sources = []
    pages_source = tool_source.parse_input_pages()
    if pages_source.inputs_defined:
        for page_source in pages_source.page_sources:
            for input_source in page_source.parse_input_sources():
                input_sources.append(input_source)
    return input_sources


class ParamContext:
    """Capture the context of a parameter's position within the inputs tree of a tool."""

    parent_context: AnyParamContext
    name: str
    # if in a repeat - what position in the repeat
    index: Optional[int]
    # we've encouraged the use of repeat/conditional tags to capture fully qualified paths
    # to parameters in tools. This brings the parameters closer to the API and prevents a
    # variety of possible ambiguities. Disable this for newer tools.
    allow_unqualified_access: bool

    def __init__(self, name: str, parent_context: AnyParamContext, index: Optional[int] = None):
        self.parent_context = parent_context
        self.name = name
        self.index = None if index is None else int(index)
        self.allow_unqualified_access = parent_context.allow_unqualified_access

    def for_state(self) -> str:
        name = self.name if self.index is None else f"{self.name}_{self.index}"
        parent_for_state = self.parent_context.for_state()
        if parent_for_state:
            return f"{parent_for_state}|{name}"
        else:
            return name

    def __str__(self) -> str:
        return f"Context[for_state={self.for_state()}]"

    def param_names(self):
        if not self.allow_unqualified_access:
            yield self.for_state()
        else:
            for parent_context_param in self.parent_context.param_names():
                if self.index is not None:
                    yield f"{parent_context_param}|{self.name}_{self.index}"
                else:
                    yield f"{parent_context_param}|{self.name}"
            if self.index is not None:
                yield f"{self.name}_{self.index}"
            else:
                yield self.name

    def extract_value(self, raw_inputs: ToolSourceTestInputs):
        for param_name in self.param_names():
            value = self.__raw_param_found(param_name, raw_inputs)
            if value:
                return value
        return None

    def __raw_param_found(self, param_name: str, raw_inputs: ToolSourceTestInputs):
        index = None
        for i, raw_input_dict in enumerate(raw_inputs):
            if raw_input_dict["name"] == param_name:
                index = i
        if index is not None:
            raw_input_dict = raw_inputs[index]
            del raw_inputs[index]
            return raw_input_dict
        else:
            return None


class RootParamContext:
    allow_unqualified_access: bool

    def __init__(self, allow_unqualified_access: bool):
        self.allow_unqualified_access = allow_unqualified_access

    def for_state(self):
        return ""

    def param_names(self):
        return []

    def get_index(self):
        return 0


def _process_simple_value(
    param: InputSource,
    param_value: Any,
    required_data_tables: RequiredDataTablesT,
    required_loc_files: RequiredLocFileT,
    allow_legacy_test_case_parameters: bool,
):
    input_type = param.get("type")
    if input_type == "select":
        # Tests may specify values as either raw value or the value
        # as they appear in the list - the API doesn't and shouldn't
        # accept the text value - so we need to convert the text
        # into the form value.
        def process_param_value(param_value):
            found_value = False
            value_for_text = None
            if allow_legacy_test_case_parameters:
                # we used to allow selections based on text - this
                # should really be only based on key
                static_options = param.parse_static_options()
                for text, opt_value, _ in static_options:
                    if param_value == opt_value:
                        found_value = True
                    if value_for_text is None and param_value == text:
                        value_for_text = opt_value
            dynamic_options = param.parse_dynamic_options()
            if dynamic_options:
                data_table_name = dynamic_options.get_data_table_name()
                index_file_name = dynamic_options.get_index_file_name()
                if data_table_name:
                    required_data_tables.append(data_table_name)
                elif index_file_name:
                    required_loc_files.append(index_file_name)
            if not found_value and value_for_text is not None:
                processed_value = value_for_text
            else:
                processed_value = param_value
            return processed_value

        # Do replacement described above for lists or singleton
        # values.
        if isinstance(param_value, list):
            processed_value = list(map(process_param_value, param_value))
        else:
            processed_value = process_param_value(param_value)
    elif input_type == "boolean":
        # Like above, tests may use the tool define values of simply
        # true/false.
        processed_value = _process_bool_param_value(param, param_value, allow_legacy_test_case_parameters)
    else:
        processed_value = param_value
    return processed_value


def _matching_case_for_value(
    tool_source: ToolSource,
    cond: InputSource,
    test_param: InputSource,
    declared_value: Any,
    allow_legacy_test_case_parameters: bool,
):
    tool_id = tool_source.parse_id()
    cond_name = cond.parse_name()

    assert test_param.parse_input_type() == "param"
    test_param_type = test_param.get("type")

    if test_param_type == "boolean":
        if declared_value is None:
            # No explicit value for param in test case, determine from default
            query_value = boolean_is_checked(test_param)
        else:
            query_value = _process_bool_param_value(test_param, declared_value, allow_legacy_test_case_parameters)

        def matches_declared_value(case_value):
            return _process_bool_param_value(test_param, case_value, allow_legacy_test_case_parameters) == query_value

    elif test_param_type == "select":
        static_options = test_param.parse_static_options()
        if declared_value is not None:
            # Test case supplied explicit value to check against.

            def matches_declared_value(case_value):
                return case_value == declared_value

        elif static_options:
            # No explicit value in test case, not much to do if options are dynamic but
            # if static options are available can find the one specified as default or
            # fallback on top most option (like GUI).

            for name, _, selected in static_options:
                if selected:
                    default_option = name
            else:
                first_option = static_options[0]
                first_option_value = first_option[1]
                default_option = first_option_value

            def matches_declared_value(case_value):
                return case_value == default_option

        else:
            # No explicit value for this param and cannot determine a
            # default - give up. Previously this would just result in a key
            # error exception.
            msg = f"Failed to find test parameter value specification required for conditional {cond_name}"
            raise Exception(msg)
    else:
        msg = f"Invalid conditional test type found {test_param_type}"
        raise Exception(msg)

    # Check the tool's defined cases against predicate to determine
    # selected or default.
    for case_when, case_input_sources in cond.parse_when_input_sources():
        if matches_declared_value(case_when):
            return case_when, case_input_sources
    else:
        msg_template = "%s - Failed to find case matching value (%s) for test parameter specification for conditional %s. Remainder of test behavior is unspecified."
        msg = msg_template % (tool_id, declared_value, cond_name)
        log.info(msg)
    return None


def _add_uploaded_dataset(
    name: str,
    value: Optional[str],
    extra: ExtraFileInfoDictT,
    input_parameter: InputSource,
    required_files: RequiredFilesT,
) -> Optional[str]:
    if value is None:
        assert (
            input_parameter.parse_optional() or "composite_data" in extra
        ), f"{name} is not optional. You must provide a valid filename."
        return value
    return require_file(name, value, extra, required_files)


def require_file(name: str, value: str, extra: ExtraFileInfoDictT, required_files: RequiredFilesT) -> str:
    if (value, extra) not in required_files:
        required_files.append((value, extra))  # these files will be uploaded
    name_changes = [att for att in extra.get("edit_attributes", []) if att.get("type") == "name"]
    if name_changes:
        name_change = name_changes[-1].get("value")  # only the last name change really matters
        value = str(name_change)  # change value for select to renamed uploaded file for e.g. composite dataset
    else:
        for end in [".zip", ".gz"]:
            if value.endswith(end):
                value = value[: -len(end)]
                break
        value = os.path.basename(value)  # if uploading a file in a path other than root of test-data
    return value


def _process_bool_param_value(
    input_source: InputSource, param_value: Any, allow_legacy_test_case_parameters: bool
) -> Any:
    truevalue, falsevalue = boolean_true_and_false_values(input_source)
    optional = input_source.parse_optional()
    return process_bool_param_value(truevalue, falsevalue, optional, param_value, allow_legacy_test_case_parameters)


def process_bool_param_value(
    truevalue: str, falsevalue: str, optional: bool, param_value: Any, allow_legacy_test_case_parameters: bool
) -> Any:
    was_list = False
    if isinstance(param_value, list):
        was_list = True
        param_value = param_value[0]

    if allow_legacy_test_case_parameters and truevalue == param_value:
        processed_value = True
    elif allow_legacy_test_case_parameters and falsevalue == param_value:
        processed_value = False
    else:
        if optional:
            processed_value = string_as_bool_or_none(param_value)
        else:
            processed_value = string_as_bool(param_value)
    return [processed_value] if was_list else processed_value


def split_if_str(value):
    split = isinstance(value, str)
    if split:
        value = value.split(",")
    return value


# convert the sort internal structure used by the tool library {tag: string, attributes: dict, children: []}
# into the YAML structure consumed by the test framework {that: string, **atributes}
def tag_structure_to_that_structure(raw_assert):
    as_json = {"that": raw_assert["tag"], **raw_assert.get("attributes", {})}
    children = raw_assert.get("children")
    if children:
        as_json["children"] = list(map(tag_structure_to_that_structure, children))
    return as_json


def assertion_xml_els_to_models(asserts_raw) -> relaxed_assertion_list:
    asserts_raw = __parse_assert_list_from_elem(asserts_raw)

    to_yaml_assertions = []
    for raw_assert in asserts_raw or []:
        to_yaml_assertions.append(tag_structure_to_that_structure(raw_assert))
    return relaxed_assertion_list.model_validate(to_yaml_assertions)



# Documented profile changes not covered.
# - Allow invalid values. @mvdbeek any clue what to say here. The PR is here:
# - https://github.com/galaxyproject/galaxy/pull/6264
#
# I don't want to declare these TODO but somethings could be more percise.
# - We could inspect the XML source and tell when tools are using galaxy.json and improve
#   the messaging on 17.09 for that. We can't be absolutely sure these would or would not be
#   problems but we can be more confident they might be.
# - Ditto for $HOME and the 18.01 migration.
# - We could try to walk parameters and give more specific advice on structured_like qualification.
from json import loads
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Type,
)

from typing_extensions import (
    Literal,
    NotRequired,
    TypedDict,
)

from galaxy.tool_util.parameters.case import validate_test_cases_for_tool_source
from galaxy.tool_util.parser.factory import get_tool_source
from galaxy.tool_util.parser.xml import XmlToolSource
from galaxy.util import Element
from galaxy.util.resources import resource_string

TOOL_TOO_NEW = "Target tool has a profile that is too new, consider upgrading this script for the latest advice."
TARGET_TOO_NEW = (
    "Target upgrade version is too new for this script, consider upgrading this script for the latest advice."
)

TYPE_LEVEL = Literal["must_fix", "consider", "ready", "info"]


class AdviceCode(TypedDict):
    name: str
    level: TYPE_LEVEL
    message: str
    niche: NotRequired[bool]
    url: NotRequired[str]


upgrade_codes_json = resource_string(__name__, "upgrade_codes.json")
upgrade_codes_by_name: Dict[str, AdviceCode] = {}

for name, upgrade_object in loads(upgrade_codes_json).items():
    upgrade_object["name"] = name
    upgrade_codes_by_name[name] = cast(AdviceCode, upgrade_object)


class Advice:
    advice_code: AdviceCode
    message: Optional[str]

    def __init__(self, advice_code: AdviceCode, message: Optional[str]):
        self.advice_code = advice_code
        self.message = message

    @property
    def url(self) -> Optional[str]:
        return self.advice_code.get("url")

    @property
    def level(self) -> TYPE_LEVEL:
        return self.advice_code["level"]

    @property
    def niche(self) -> bool:
        return self.advice_code.get("niche", False)

    @property
    def advice_code_message(self) -> str:
        return self.advice_code["message"]

    def to_dict(self) -> Dict[str, Any]:
        as_dict = cast(Dict[str, Any], self.advice_code.copy())
        as_dict["advice_code_message"] = self.advice_code_message
        as_dict["message"] = self.message
        return as_dict


class AdviceCollection:
    _advice: List[Advice]

    def __init__(self):
        self._advice = []

    def add(self, code: str, message: Optional[str] = None):
        self._advice.append(Advice(upgrade_codes_by_name[code], message))

    def to_list(self) -> List[Advice]:
        return self._advice


class ProfileMigration:
    """A class offering advice on upgrading a Galaxy tool between two profile versions."""

    from_version: str
    to_version: str

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        return None


class ProfileMigration16_04(ProfileMigration):
    from_version = "16.01"
    to_version = "16.04"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        interpreter = tool_source.parse_interpreter()
        if interpreter:
            advice_collection.add("16_04_fix_interpreter")
        else:
            advice_collection.add("16_04_ready_interpreter")
        advice_collection.add("16_04_consider_implicit_extra_file_collection")

        if _has_matching_xpath(tool_source, ".//data[@format = 'input']"):
            advice_collection.add("16_04_fix_output_format")

        if not _has_matching_xpath(tool_source, ".//stdio") and not _has_matching_xpath(
            tool_source, ".//command[@detect_errors]"
        ):
            advice_collection.add("16_04_exit_code")


class ProfileMigration17_09(ProfileMigration):
    from_version = "16.04"
    to_version = "17.09"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)

        outputs_el = tool_source._outputs_el
        if outputs_el is not None and outputs_el.get("`provided_metadata_style`", None) is not None:
            advice_collection.add("17_09_consider_provided_metadata_style")


class ProfileMigration18_01(ProfileMigration):
    from_version = "17.09"
    to_version = "18.01"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        command_el = tool_source._command_el
        if command_el is not None and command_el.get("use_shared_home", None) is None:
            advice_collection.add("18_01_consider_home_directory")

        if _has_matching_xpath(tool_source, ".//outputs/collection[@structured_like]"):
            advice_collection.add("18_01_consider_structured_like")


class ProfileMigration18_09(ProfileMigration):
    from_version = "18.01"
    to_version = "18.09"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        tool_type = tool_source.parse_tool_type()
        if tool_type == "manage_data":
            advice_collection.add("18_09_consider_python_environment")


class ProfileMigration20_05(ProfileMigration):
    from_version = "18.09"
    to_version = "20.05"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)

        if _has_matching_xpath(tool_source, ".//configfiles/inputs"):
            advice_collection.add("20_05_consider_inputs_as_json_changes")


class ProfileMigration20_09(ProfileMigration):
    from_version = "20.05"
    to_version = "20.09"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)

        tests = tool_source.parse_tests_to_dict()
        for test in tests["tests"]:
            output_collections = test.get("output_collections")
            if not output_collections:
                continue

            for output_collection in output_collections:
                if output_collection.get("element_tests"):
                    advice_collection.add("20_09_consider_output_collection_order")

        command_el = tool_source._command_el
        if command_el is not None:
            strict = command_el.get("strict", None)
            if strict is None:
                advice_collection.add("20_09_consider_set_e")


class ProfileMigration21_09(ProfileMigration):
    from_version = "20.09"
    to_version = "21.09"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        for el in _find_all(tool_source, ".//data[@from_work_dir]"):
            from_work_dir = el.get("from_work_dir") or ""
            if from_work_dir != from_work_dir.strip():
                advice_collection.add("")

        tool_type = tool_source.parse_tool_type()
        if tool_type == "data_source":
            advice_collection.add("21_09_consider_python_environment")


class ProfileMigration23_0(ProfileMigration):
    from_version = "21.09"
    to_version = "23.0"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        for text_param in _find_all(tool_source, ".//input[@type='text']"):
            optional_tag_set = text_param.get("optional", None) is not None
            if not optional_tag_set:
                advice_collection.add("23_0_consider_optional_text")


class ProfileMigration24_0(ProfileMigration):
    from_version = "23.0"
    to_version = "24.0"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        tool_type = tool_source.parse_tool_type()
        if tool_type == "data_source_async":
            advice_collection.add("24_0_consider_python_environment")
        if tool_type in ["data_source_async", "data_source"]:
            advice_collection.add("24_0_request_cleaning")


class ProfileMigration24_2(ProfileMigration):
    from_version = "24.0"
    to_version = "24.2"

    @classmethod
    def advise(cls, advice_collection: AdviceCollection, xml_file: str) -> None:
        tool_source = _xml_tool_source(xml_file)
        results = validate_test_cases_for_tool_source(tool_source, use_latest_profile=True)
        for result in results:
            if result.validation_error:
                advice_collection.add("24_2_fix_test_case_validation", str(result.validation_error))


profile_migrations: List[Type[ProfileMigration]] = [
    ProfileMigration16_04,
    ProfileMigration17_09,
    ProfileMigration18_01,
    ProfileMigration18_09,
    ProfileMigration20_05,
    ProfileMigration20_09,
    ProfileMigration21_09,
    ProfileMigration23_0,
    ProfileMigration24_0,
    ProfileMigration24_2,
]

latest_supported_version = "24.2"


def advise_on_upgrade(xml_file: str, to_version: Optional[str] = None) -> List[Advice]:
    to_version = to_version or latest_supported_version
    tool_source = _xml_tool_source(xml_file)
    initial_version = tool_source.parse_profile()
    if initial_version > latest_supported_version:
        raise Exception(TOOL_TOO_NEW)
    elif to_version > latest_supported_version:
        raise Exception(TARGET_TOO_NEW)
    advice_collection = AdviceCollection()

    for migration in profile_migrations:
        if migration.to_version < initial_version:
            # tool started past here... just skip this advice
            continue
        if migration.to_version > to_version:
            # we're not advising on upgrading past this point
            break
        migration.advise(advice_collection, xml_file)

    return advice_collection.to_list()


def _xml_tool_source(xml_file: str) -> XmlToolSource:
    tool_source = get_tool_source(xml_file)
    if not isinstance(tool_source, XmlToolSource):
        raise Exception("Can only provide upgrade advice for XML tools.")
    return tool_source


def _has_matching_xpath(tool_source: XmlToolSource, xpath: str) -> bool:
    return tool_source.xml_tree.find(xpath) is not None


def _find_all(tool_source: XmlToolSource, xpath: str) -> List[Element]:
    return cast(List[Element], tool_source.xml_tree.findall(".//data[@from_work_dir]") or [])



import os.path
from typing import TYPE_CHECKING

from galaxy import tool_util
from galaxy.tool_util.lint import Linter
from galaxy.util import etree

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser import ToolSource

TOOL_XSD = os.path.join(os.path.dirname(tool_util.__file__), "xsd", "galaxy.xsd")


class XSD(Linter):
    """
    Lint an XML tool against XSD
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        xmlschema = etree.parse(TOOL_XSD)
        xsd = etree.XMLSchema(xmlschema)
        if not xsd.validate(tool_xml):
            for error in xsd.error_log:  # type: ignore[attr-defined] # https://github.com/lxml/lxml-stubs/pull/100
                # the validation error contains the path which allows
                # us to lookup the node that is reported in the lint context
                node = tool_xml.xpath(error.path)
                node = node[0]
                lint_ctx.error(f"Invalid XML: {error.message}", linter=cls.name(), node=node)



import json
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
)

from typing_extensions import get_args

from galaxy.tool_util_models.parameter_validators import (
    AnyValidatorModel,
    DatasetMetadataEqualParameterValidatorModel,
    DatasetMetadataInDataTableParameterValidatorModel,
    DatasetMetadataInFileParameterValidatorModel,
    DatasetMetadataInRangeParameterValidatorModel,
    DatasetMetadataNotInDataTableParameterValidatorModel,
    DatasetOkValidatorParameterValidatorModel,
    DiscriminatedAnyValidatorModel,
    EmptyDatasetParameterValidatorModel,
    EmptyExtraFilesPathParameterValidatorModel,
    EmptyFieldParameterValidatorModel,
    ExpressionParameterValidatorModel,
    InRangeParameterValidatorModel,
    LengthParameterValidatorModel,
    MetadataParameterValidatorModel,
    NoOptionsParameterValidatorModel,
    ParameterValidatorModel,
    RegexParameterValidatorModel,
    SPLIT_DEFAULT,
    StaticValidatorModel,
    UnspecifiedBuildParameterValidatorModel,
    ValidatorType,
    ValueInDataTableParameterValidatorModel,
    ValueNotInDataTableParameterValidatorModel,
)
from galaxy.util import (
    asbool,
    Element,
)


class UnsafeValidatorConfiguredInUntrustedContext(AssertionError):
    pass


def parse_dict_validators(validator_dicts: List[Dict[str, Any]], trusted: bool) -> List[AnyValidatorModel]:
    validator_models = []
    for validator_dict in validator_dicts:
        validator = DiscriminatedAnyValidatorModel.validate_python(validator_dict)
        if not trusted:
            # Don't risk instantiating unsafe validators for user-defined code
            if not validator._safe:
                raise UnsafeValidatorConfiguredInUntrustedContext()
        validator_models.append(validator)
    return validator_models


def parse_xml_validators(input_elem: Element) -> List[AnyValidatorModel]:
    validator_els: List[Element] = input_elem.findall("validator") or []
    models = []
    for validator_el in validator_els:
        models.append(parse_xml_validator(validator_el))
    return models


def static_validators(validator_models: List[AnyValidatorModel]) -> List[AnyValidatorModel]:
    static_validators = []
    for validator_model in validator_models:
        if validator_model._static:
            static_validators.append(validator_model)
    return static_validators


def parse_xml_validator(validator_el: Element) -> AnyValidatorModel:
    _type = validator_el.get("type")
    if _type is None:
        raise ValueError("Required 'type' attribute missing from validator")
    valid_types = get_args(ValidatorType)
    if _type not in valid_types:
        raise ValueError(f"Unknown 'type' attribute in validator {_type}")
    validator_type: ValidatorType = cast(ValidatorType, _type)
    if validator_type == "expression":
        return ExpressionParameterValidatorModel(
            type="expression",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
            expression=validator_el.text,
        )
    elif validator_type == "regex":
        return RegexParameterValidatorModel(
            type="regex",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
            expression=validator_el.text,
        )
    elif validator_type == "in_range":
        return InRangeParameterValidatorModel(
            type="in_range",
            message=_parse_message(validator_el),
            min=_parse_number(validator_el, "min"),
            max=_parse_number(validator_el, "max"),
            exclude_min=_parse_bool(validator_el, "exclude_min", False),
            exclude_max=_parse_bool(validator_el, "exclude_max", False),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "length":
        return LengthParameterValidatorModel(
            type="length",
            min=_parse_int(validator_el, "min"),
            max=_parse_int(validator_el, "max"),
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "metadata":
        return MetadataParameterValidatorModel(
            type="metadata",
            message=_parse_message(validator_el),
            check=_parse_str_list(validator_el, "check"),
            skip=_parse_str_list(validator_el, "skip"),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "dataset_metadata_equal":
        return DatasetMetadataEqualParameterValidatorModel(
            type="dataset_metadata_equal",
            metadata_name=validator_el.get("metadata_name"),
            value=_parse_json_value(validator_el),
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "unspecified_build":
        return UnspecifiedBuildParameterValidatorModel(
            type="unspecified_build",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "no_options":
        return NoOptionsParameterValidatorModel(
            type="no_options",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "empty_field":
        return EmptyFieldParameterValidatorModel(
            type="empty_field",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "empty_dataset":
        return EmptyDatasetParameterValidatorModel(
            type="empty_dataset",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "empty_extra_files_path":
        return EmptyExtraFilesPathParameterValidatorModel(
            type="empty_extra_files_path",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "dataset_metadata_in_data_table":
        return DatasetMetadataInDataTableParameterValidatorModel(
            type="dataset_metadata_in_data_table",
            message=_parse_message(validator_el),
            table_name=validator_el.get("table_name"),
            metadata_name=validator_el.get("metadata_name"),
            metadata_column=_parse_metadata_column(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "dataset_metadata_not_in_data_table":
        return DatasetMetadataNotInDataTableParameterValidatorModel(
            type="dataset_metadata_not_in_data_table",
            message=_parse_message(validator_el),
            table_name=validator_el.get("table_name"),
            metadata_name=validator_el.get("metadata_name"),
            metadata_column=_parse_metadata_column(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "dataset_metadata_in_range":
        return DatasetMetadataInRangeParameterValidatorModel(
            type="dataset_metadata_in_range",
            message=_parse_message(validator_el),
            metadata_name=validator_el.get("metadata_name"),
            min=_parse_number(validator_el, "min"),
            max=_parse_number(validator_el, "max"),
            exclude_min=_parse_bool(validator_el, "exclude_min", False),
            exclude_max=_parse_bool(validator_el, "exclude_max", False),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "value_in_data_table":
        return ValueInDataTableParameterValidatorModel(
            type="value_in_data_table",
            message=_parse_message(validator_el),
            table_name=validator_el.get("table_name"),
            metadata_column=_parse_metadata_column(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "value_not_in_data_table":
        return ValueNotInDataTableParameterValidatorModel(
            type="value_not_in_data_table",
            message=_parse_message(validator_el),
            table_name=validator_el.get("table_name"),
            metadata_column=_parse_metadata_column(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "dataset_ok_validator":
        return DatasetOkValidatorParameterValidatorModel(
            type="dataset_ok_validator",
            message=_parse_message(validator_el),
            negate=_parse_negate(validator_el),
        )
    elif validator_type == "dataset_metadata_in_file":
        filename = validator_el.get("filename")
        return DatasetMetadataInFileParameterValidatorModel(
            type="dataset_metadata_in_file",
            message=_parse_message(validator_el),
            filename=filename,
            metadata_name=validator_el.get("metadata_name"),
            metadata_column=_parse_metadata_column(validator_el),
            line_startswith=validator_el.get("line_startswith"),
            split=validator_el.get("split", SPLIT_DEFAULT),
            negate=_parse_negate(validator_el),
        )
    else:
        raise ValueError(f"Unhandled 'type' attribute in validator {validator_type}")


def _parse_message(xml_el: Element) -> Optional[str]:
    message = xml_el.get("message")
    return message


def _parse_int(xml_el: Element, attribute: str) -> Optional[int]:
    raw_value = xml_el.get(attribute)
    if raw_value:
        return int(raw_value)
    else:
        return None


def _parse_number(xml_el: Element, attribute: str) -> Optional[Union[float, int]]:
    raw_value = xml_el.get(attribute)
    if raw_value and ("." in raw_value or "e" in raw_value or "inf" in raw_value):
        return float(raw_value)
    elif raw_value:
        return int(raw_value)
    else:
        return None


def _parse_negate(xml_el: Element) -> bool:
    return _parse_bool(xml_el, "negate", False)


def _parse_bool(xml_el: Element, attribute: str, default_value: bool) -> bool:
    return asbool(xml_el.get(attribute, default_value))


def _parse_str_list(xml_el: Element, attribute: str) -> List[str]:
    raw_value = xml_el.get(attribute)
    if not raw_value:
        return []
    else:
        return [v.strip() for v in raw_value.split(",")]


def _parse_json_value(xml_el: Element) -> Any:
    value = xml_el.get("value", None) or json.loads(xml_el.get("value_json", "null"))
    return value


def _parse_metadata_column(xml_el: Element) -> Union[int, str]:
    column = xml_el.get("metadata_column", 0)
    try:
        return int(column)
    except ValueError:
        return column


def static_tool_validators(validators: Sequence[ParameterValidatorModel]) -> List[StaticValidatorModel]:
    static_validators: List[StaticValidatorModel] = []
    for validator in validators:
        if isinstance(validator, StaticValidatorModel):
            static_validators.append(validator)
    return static_validators


def statically_validates(validators: Sequence[ParameterValidatorModel], value: Any) -> bool:
    for validator in static_tool_validators(validators):
        try:
            validator.statically_validate(value)
        except ValueError:
            return False
    return True



from typing import (
    Dict,
    Optional,
    TYPE_CHECKING,
)

from galaxy.util.tool_version import remove_version_from_guid
from .interface import ToolLineage

if TYPE_CHECKING:
    from galaxy.tools import Tool


class LineageMap:
    """Map each unique tool id to a lineage object."""

    def __init__(self, app):
        self.lineage_map: Dict[str, ToolLineage] = {}
        self.app = app

    def register(self, tool: "Tool") -> ToolLineage:
        tool_id = tool.id
        assert tool_id
        versionless_tool_id = remove_version_from_guid(tool_id)
        lineage: ToolLineage
        if versionless_tool_id not in self.lineage_map:
            lineage = ToolLineage.from_tool(tool)
        else:
            lineage = self.lineage_map[versionless_tool_id]
            # A lineage for a tool with the same versionless_tool_id exists,
            # but this lineage may not have the current tools' version,
            # so we add tool.version to the lineage
            lineage.register_version(tool.version)
        if versionless_tool_id and versionless_tool_id not in self.lineage_map:
            self.lineage_map[versionless_tool_id] = lineage
        if tool_id not in self.lineage_map:
            self.lineage_map[tool_id] = lineage
        return self.lineage_map[tool_id]

    def get(self, tool_id: str) -> Optional[ToolLineage]:
        """
        Get lineage for `tool_id`.

        By preference the lineage for a version-agnostic tool_id is returned.
        Falls back to fetching the lineage only when this fails.
        This happens when the tool_id does not contain a version.
        """
        lineage = self._get_versionless(tool_id)
        if lineage:
            return lineage
        if tool_id not in self.lineage_map:
            toolbox = None
            try:
                toolbox = self.app.toolbox
            except AttributeError:
                # We're building the lineage map while building the toolbox,
                # so app.toolbox may not be available.
                # TODO: is the fallback really needed / can it be fixed by improving _get_versionless ?
                pass
            tool = toolbox and toolbox._tools_by_id.get(tool_id)
            if tool:
                lineage = ToolLineage.from_tool(tool)
                self.lineage_map[tool_id] = lineage
        return self.lineage_map.get(tool_id)

    def _get_versionless(self, tool_id: str) -> Optional[ToolLineage]:
        versionless_tool_id = remove_version_from_guid(tool_id)
        if not versionless_tool_id:
            return None
        return self.lineage_map.get(versionless_tool_id)


__all__ = ("LineageMap",)



import json
from typing import (
    Any,
    Dict,
)

from pydantic.json_schema import GenerateJsonSchema
from typing_extensions import Literal

MODE = Literal["validation", "serialization"]
DEFAULT_JSON_SCHEMA_MODE: MODE = "validation"


class CustomGenerateJsonSchema(GenerateJsonSchema):

    def generate(self, schema, mode: MODE = DEFAULT_JSON_SCHEMA_MODE):
        json_schema = super().generate(schema, mode=mode)
        json_schema["$schema"] = self.schema_dialect
        return json_schema


def to_json_schema(model, mode: MODE = DEFAULT_JSON_SCHEMA_MODE) -> Dict[str, Any]:
    return model.model_json_schema(schema_generator=CustomGenerateJsonSchema, mode=mode)


def to_json_schema_string(model, mode: MODE = DEFAULT_JSON_SCHEMA_MODE) -> str:
    return json.dumps(to_json_schema(model, mode=mode), indent=4)



from collections import OrderedDict
from typing import (
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
)

from packaging.version import Version

from galaxy.util import string_as_bool
from .parameter_validators import statically_validates

if TYPE_CHECKING:
    from .interface import (
        InputSource,
        ToolSource,
    )

DEFAULT_DELTA = 10000
DEFAULT_DELTA_FRAC = None

DEFAULT_METRIC = "mae"
DEFAULT_EPS = 0.01
DEFAULT_PIN_LABELS = None
DEFAULT_SORT = False
DEFAULT_DECOMPRESS = False


def is_dict(item):
    return isinstance(item, dict) or isinstance(item, OrderedDict)


def _parse_name(name, argument):
    """Determine name of an input source from name and argument
    returns the name or if absent the argument property
    In the latter case, leading dashes are stripped and
    all remaining dashes are replaced by underscores.
    """
    if name is None:
        if argument is None:
            raise ValueError("parameter must specify a 'name' or 'argument'.")
        name = argument.lstrip("-").replace("-", "_")
    return name


def parse_profile_version(tool_source: "ToolSource") -> float:
    return float(tool_source.parse_profile())


def parse_tool_version_with_defaults(
    id: Optional[str], tool_source: "ToolSource", profile: Optional[Version] = None
) -> str:
    if profile is None:
        profile = Version(tool_source.parse_profile())

    version = tool_source.parse_version()
    if not version:
        if profile < Version("16.04"):
            # For backward compatibility, some tools may not have versions yet.
            version = "1.0.0"
        else:
            raise ValueError(f"Missing tool 'version' for tool with id '{id}' at '{tool_source}'")
    return version


def boolean_is_checked(input_source: "InputSource"):
    nullable = input_source.get_bool("optional", False)
    return input_source.get_bool("checked", None if nullable else False)


def boolean_true_and_false_values(input_source, profile: Optional[Union[float, str]] = None) -> Tuple[str, str]:
    truevalue = input_source.get("truevalue", "true")
    falsevalue = input_source.get("falsevalue", "false")
    if profile and Version(str(profile)) >= Version("23.1"):
        if truevalue == falsevalue:
            raise ParameterParseException("Cannot set true and false to the same value")
        if truevalue.lower() == "false":
            raise ParameterParseException(
                f"Cannot set truevalue to [{truevalue}], Galaxy state may encounter issues distinguishing booleans and strings in this case."
            )
        if falsevalue.lower() == "true":
            raise ParameterParseException(
                f"Cannot set falsevalue to [{falsevalue}], Galaxy state may encounter issues distinguishing booleans and strings in this case."
            )
    return (truevalue, falsevalue)


def text_input_is_optional(input_source: "InputSource") -> Tuple[bool, bool]:
    # Optionality not explicitly defined, default to False
    optional: Optional[bool] = False
    optionality_inferred: bool = False

    optional = input_source.get("optional", None)
    if optional is not None:
        optional = string_as_bool(optional)
    else:
        # A text parameter that doesn't raise a validation error on empty string
        # is considered to be optional
        if statically_validates(input_source.parse_validators(), ""):
            optional = True
            optionality_inferred = True
        else:
            optional = False

    assert isinstance(optional, bool)
    return optional, optionality_inferred


class ParameterParseException(Exception):
    message: str

    def __init__(self, message):
        super().__init__(message)
        self.message = message


def multiple_select_value_split(values: Union[str, List[str]]) -> List[str]:
    # used to split simple strings into lists from both tool XML and from the API for consistency
    value_list = []
    if not isinstance(values, list):
        values = values.split("\n")
    for value in values:
        for value_split in str(value).split(","):
            value_split = value_split.strip()
            if value_split:
                value_list.append(value_split)
    return value_list



from typing import (
    Any,
    Dict,
    Optional,
    Type,
)

from pydantic import (
    BaseModel,
    ValidationError,
)
from typing_extensions import (
    Protocol,
)

from galaxy.exceptions import RequestParameterInvalidException
from galaxy.tool_util_models.parameters import (
    create_field_model,
    DEFAULT_MODEL_NAME,
    RawStateDict,
    StateRepresentationT,
    ToolParameterBundle,
)


def validate_against_model(pydantic_model: Type[BaseModel], parameter_state: Dict[str, Any]) -> None:
    try:
        pydantic_model(**parameter_state)
    except ValidationError as e:
        # TODO: Improve this or maybe add a handler for this in the FastAPI exception
        # handler.
        raise RequestParameterInvalidException(str(e))


class ValidationFunctionT(Protocol):

    def __call__(self, tool: ToolParameterBundle, request: RawStateDict, name: Optional[str] = None) -> None: ...


def validate_model_type_factory(state_representation: StateRepresentationT) -> ValidationFunctionT:

    def validate_request(tool: ToolParameterBundle, request: Dict[str, Any], name: Optional[str] = None) -> None:
        name = name or DEFAULT_MODEL_NAME
        pydantic_model = create_field_model(tool.parameters, name=name, state_representation=state_representation)
        validate_against_model(pydantic_model, request)

    return validate_request


validate_relaxed_request = validate_model_type_factory("relaxed_request")
validate_request = validate_model_type_factory("request")
validate_internal_request = validate_model_type_factory("request_internal")
validate_internal_request_dereferenced = validate_model_type_factory("request_internal_dereferenced")
validate_landing_request = validate_model_type_factory("landing_request")
validate_internal_landing_request = validate_model_type_factory("landing_request_internal")
validate_internal_job = validate_model_type_factory("job_internal")
validate_test_case = validate_model_type_factory("test_case_xml")
validate_workflow_step = validate_model_type_factory("workflow_step")
validate_workflow_step_linked = validate_model_type_factory("workflow_step_linked")



import argparse
import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import (
    List,
    Optional,
)

from galaxy.tool_util.parameters.case import (
    TestCaseStateValidationResult,
    validate_test_cases_for_tool_source,
)
from galaxy.tool_util.parser import get_tool_source

DESCRIPTION = """
A small utility to load a Galaxy tool file and run tool test validation code on it.

This is not meant to be an end-user application or part of the typical tool developer tool
chain. This functionality will be integrated into Planemo with the typical polish that
entails. This purpose of this utility is for Galaxy developers and people wishing to
make informed decisions about the tool test schema and best practices to apply the tool
state validation code in isolation. This script can also be used to integrate this functionality
in non-Python tool chains.
"""


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("tool_source")
    parser.add_argument(
        "-l",
        "--latest",
        dest="latest",
        default=False,
        action="store_true",
        help="Validate test against latest tool profile version regardless of explicit tool profile version specified in tool source file.",
    )
    parser.add_argument(
        "-j",
        "--json",
        default=False,
        action="store_true",
        help="Output validation results as JSON.",
    )
    return parser


@dataclass
class ToolTestValidationResults:
    tool_id: str
    tool_version: Optional[str]
    tool_profile: str
    tool_path: str
    results: List[TestCaseStateValidationResult]
    load_error: Optional[Exception]

    def to_dict(self):
        return {
            "tool_id": self.tool_id,
            "tool_version": self.tool_version,
            "tool_path": str(self.tool_path),
            "results": [r.to_dict() for r in self.results],
            "load_error": str(self.load_error) if self.load_error else None,
        }


def report_results(results_for_tool: ToolTestValidationResults) -> None:
    test_results = results_for_tool.results
    print(
        f"Found {len(test_results)} test cases to validate for tool {results_for_tool.tool_id} / {results_for_tool.tool_version} (@ {results_for_tool.tool_path})"
    )
    for i, result in enumerate(test_results):
        if result.validation_error is not None:
            print(f"Test Case {i + 1}: validation failed - {str(result.validation_error)}")
        elif len(result.warnings) == 0:
            print(f"Test Case {i + 1}: validated")
        else:
            print(f"Test case {i + 1}: test case validated with warnings:")
            for test_warning in result.warnings:
                print(f"- {test_warning}")


class NotValidToolException(Exception):
    pass


def validate_tool(tool_path, latest) -> ToolTestValidationResults:
    try:
        tool_source = get_tool_source(tool_path)
    except Exception:
        # probably not a Galaxy tool... just skip
        raise NotValidToolException()
    tool_id = tool_source.parse_id()
    if tool_id is None:
        raise NotValidToolException()
    load_error: Optional[Exception] = None
    try:
        results = validate_test_cases_for_tool_source(tool_source, use_latest_profile=latest)
    except Exception as e:
        load_error = e
        results = []
    return ToolTestValidationResults(
        tool_id,
        tool_source.parse_version(),
        tool_source.parse_profile(),
        tool_path,
        results,
        load_error,
    )


def validate_test_cases(args):
    path = args.tool_source
    latest = args.latest
    if not os.path.isdir(path):
        results = validate_tool(path, latest)
        if args.json:
            print(json.dumps(results.to_dict()))
        else:
            report_results(results)
    else:
        xml_files = Path(path).glob("**/*.xml")
        all_results = []
        for xml_file in xml_files:
            if "test-data" in str(xml_file):
                continue
            try:
                results = validate_tool(xml_file, latest)
                all_results.append(results)
            except NotValidToolException:
                continue

        if args.json:
            print(json.dumps([r.to_dict() for r in all_results], indent=4))
        else:
            for results in all_results:
                report_results(results)


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv[1:]

    args = arg_parser().parse_args(argv)
    validate_test_cases(args)


if __name__ == "__main__":
    main()



"""This module contains a linting functions for tool error detection."""

import re
from typing import TYPE_CHECKING

from packaging.version import Version

from galaxy.tool_util.lint import Linter

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource


class StdIOAbsenceLegacy(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            # Can only lint XML tools at this point.
            # Should probably use tool_source.parse_stdio() to abstract away XML details
            return
        stdios = tool_xml.findall("./stdio") if tool_xml else []
        if stdios:
            return
        tool_node = tool_xml.getroot()
        command = tool_xml.find("./command")
        if command is None or not command.get("detect_errors"):
            if Version(tool_source.parse_profile()) <= Version("16.01"):
                lint_ctx.info(
                    "No stdio definition found, tool indicates error conditions with output written to stderr.",
                    linter=cls.name(),
                    node=tool_node,
                )


class StdIOAbsence(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            # Can only lint XML tools at this point.
            # Should probably use tool_source.parse_stdio() to abstract away XML details
            return
        stdios = tool_xml.findall("./stdio") if tool_xml else []
        if stdios:
            return
        tool_node = tool_xml.getroot()
        command = tool_xml.find("./command")
        if command is None or not command.get("detect_errors"):
            if Version(tool_source.parse_profile()) > Version("16.01"):
                lint_ctx.info(
                    "No stdio definition found, tool indicates error conditions with non-zero exit codes.",
                    linter=cls.name(),
                    node=tool_node,
                )


class StdIORegex(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            # Can only lint XML tools at this point.
            # Should probably use tool_source.parse_stdio() to abstract away XML details
            return
        stdios = tool_xml.findall("./stdio") if tool_xml else []

        if len(stdios) != 1:
            return

        stdio = stdios[0]
        for child in list(stdio):
            if child.tag == "regex":
                match = child.attrib.get("match")
                if match:
                    try:
                        re.compile(match)
                    except Exception as e:
                        lint_ctx.error(
                            f"Match '{match}' is no valid regular expression: {str(e)}", linter=cls.name(), node=child
                        )



import os
from typing import (
    Dict,
    Type,
    TYPE_CHECKING,
)

from galaxy.util import plugin_config

if TYPE_CHECKING:
    from galaxy.tool_util.locations import ToolLocationResolver
    from galaxy.util.path import StrPath


class ToolLocationFetcher:
    def __init__(self):
        self.resolver_classes = self.__resolvers_dict()

    def __resolvers_dict(self) -> Dict[str, Type["ToolLocationResolver"]]:
        import galaxy.tool_util.locations

        return plugin_config.plugins_dict(galaxy.tool_util.locations, "scheme")

    def to_tool_path(self, path_or_uri_like: "StrPath", **kwds) -> "StrPath":
        if isinstance(path_or_uri_like, os.PathLike) or "://" not in path_or_uri_like:
            path = path_or_uri_like
        else:
            uri_like = path_or_uri_like
            if ":" not in path_or_uri_like:
                raise Exception("Invalid URI passed to get_tool_source")
            scheme = uri_like.split(":", 2)[0]
            if scheme not in self.resolver_classes:
                raise Exception(f"Unknown tool scheme [{scheme}] for URI [{uri_like}]")
            path = self.resolver_classes[scheme]().get_tool_source_path(uri_like)

        return path
